# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""KQL Driver class."""
from datetime import datetime
import re
from typing import Tuple, Union, Any, Dict, Optional, Iterable

import json
import pandas as pd
from IPython import get_ipython
from adal.adal_error import AdalError
from Kqlmagic.kql_response import KqlError
from Kqlmagic.kql_engine import KqlEngineError
from Kqlmagic.my_aad_helper import AuthenticationError

from .driver_base import DriverBase, QuerySource
from ...common.exceptions import (
    MsticpyNoDataSourceError,
    MsticpyNotConnectedError,
    MsticpyKqlConnectionError,
    MsticpyDataQueryError,
)
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class KqlDriver(DriverBase):
    """KqlDriver class to execute kql queries."""

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiaite KqlDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        """
        self._ip = get_ipython()
        self._debug = kwargs.get("debug", False)
        super().__init__()

        self.formatters = {"datetime": self._format_datetime, "list": self._format_list}
        self._loaded = self._is_kqlmagic_loaded()

        if not self._loaded:
            self._load_kql_magic()

        self._schema: Dict[str, Any] = {}

        if connection_str:
            self.current_connection = connection_str
            self.connect(connection_str)

    # pylint: disable=too-many-branches
    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connect to a data source

        """
        if not connection_str:
            raise MsticpyKqlConnectionError(
                "A connection string is needed to connect to Azure Sentinel.",
                title="no connection string",
            )
        self.current_connection = connection_str
        kql_err_setting = self._get_kql_option("Kqlmagic.short_errors")
        self._connected = False
        try:
            self._set_kql_option("Kqlmagic.short_errors", False)
            if self._ip is not None:
                try:
                    self._ip.run_cell_magic("kql", line="", cell=connection_str)
                except KqlError as ex:
                    self._raise_kql_error(ex)
                except KqlEngineError as ex:
                    self._raise_kql_engine_error(ex)
                except AdalError as ex:
                    self._raise_adal_error(ex)
                except AuthenticationError as ex:
                    self._raise_authn_error(ex)
                self._connected = True
                self._schema = self._get_schema()
            else:
                print(f"Could not connect to kql query provider for {connection_str}")
            return self._connected
        finally:
            self._set_kql_option("Kqlmagic.short_errors", kql_err_setting)

    # pylint: disable=too-many-branches

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._schema

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        del kwargs
        if query_source:
            try:
                table = query_source["args.table"]
            except KeyError:
                table = None
            if table:
                if " " in table.strip():
                    table = table.strip().split(" ")[0]
                if table not in self.schema:
                    raise MsticpyNoDataSourceError(
                        f"The table {table} for this query is not in your workspace",
                        " schema. Please check your workspace",
                        title=f"{table} not found.",
                    )
        data, result = self.query_with_results(query)
        return data if data is not None else result

    # pylint: disable=too-many-branches
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) and
            Kql ResultSet.

        """
        # connect or switch the connection if our connection string
        # is not the current KqlMagic connection.
        if not self.connected and self.current_connection:
            self.connect(self.current_connection)
        if not self.connected:
            raise MsticpyNotConnectedError(
                "Please run the connect() method before running a query.",
                title="not connected to a workspace.",
                help_uri=MsticpyKqlConnectionError.DEF_HELP_URI,
            )

        if self._debug:
            print(query)

        # save current auto_dataframe setting so that we can set to false
        # and restore current setting
        auto_dataframe = self._get_kql_option(option="Kqlmagic.auto_dataframe")
        self._set_kql_option(option="Kqlmagic.auto_dataframe", value=False)
        # run the query (append semicolon to prevent default output)
        if not query.strip().endswith(";"):
            query = f"{query}\n;"

        result = self._ip.run_cell_magic("kql", line="", cell=query)
        self._set_kql_option(option="Kqlmagic.auto_dataframe", value=auto_dataframe)
        if result is not None:
            if isinstance(result, pd.DataFrame):
                return result, None
            if (
                hasattr(result, "completion_query_info")
                and result.completion_query_info["StatusCode"] == 0
            ):
                data_frame = result.to_dataframe()
                if result.is_partial_table:
                    print("Warning - query returned partial results.")
                return data_frame, result

        # Query failed
        err_args = []
        if hasattr(result, "completion_query_info"):
            err_desc = result.completion_query_info.get("StatusDescription")
            err_desc = f"StatusDescription {err_desc}"
            err_code = f"(err_code: {result.completion_query_info.get('StatusCode')})"
            err_args = [err_desc, err_code]
        err_args.append(f"Query:\n{query}")
        raise MsticpyDataQueryError(*err_args)

    def _load_kql_magic(self):
        """Load KqlMagic if not loaded."""
        # KqlMagic
        print("Please wait. Loading Kqlmagic extension...")
        if self._ip is not None:
            self._ip.run_line_magic("reload_ext", "Kqlmagic")
            self._loaded = True

    def _is_kqlmagic_loaded(self) -> bool:
        """Return true if kql magic is loaded."""
        if self._ip is not None:
            return self._ip.find_magic("kql") is not None
        return False

    def _get_schema(self) -> Dict[str, Dict]:
        return self._ip.run_line_magic("kql", line="--schema")

    def _get_kql_option(self, option):
        """Retrieve a current Kqlmagic notebook option."""
        if self._ip is None:
            return None
        return self._ip.run_line_magic("config", line=option)

    def _set_kql_option(self, option, value):
        """Set a Kqlmagic notebook option."""
        if self._ip is None:
            return None
        set_txt = f"{option}={value}"
        return self._ip.run_line_magic("config", line=set_txt)

    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return date_time.isoformat(sep="T") + "Z"

    @staticmethod
    def _format_list(param_list: Iterable[Any]):
        """Return formatted list parameter."""
        fmt_list = []
        for item in param_list:
            if isinstance(item, str):
                fmt_list.append(f"'{item}'")
            else:
                fmt_list.append(f"{item}")
        return ",".join(fmt_list)

    _WS_RGX = r"workspace\(['\"](?P<ws>[^'\"]+)"
    _TEN_RGX = r"tenant\(['\"](?P<tenant>[^'\"]+)"

    def _raise_kql_error(self, ex):
        kql_err = json.loads(ex.args[0]).get("error")
        if kql_err.get("code") == "WorkspaceNotFoundError":
            ex_mssgs = [
                "The workspace ID used to connect to Azure Sentinel could not be found.",
                "Please check that this is a valid workspace for your subscription",
            ]
            ws_match = re.search(self._WS_RGX, self.current_connection, re.IGNORECASE)
            if ws_match:
                ws_name = ws_match.groupdict().get("ws")
                ex_mssgs.append(f"The workspace id used was {ws_name}.")
            ex_mssgs.append(f"The full connection string was {self.current_connection}")
            raise MsticpyKqlConnectionError(*ex_mssgs, title="unknown workspace")
        raise MsticpyKqlConnectionError(
            "The service returned the following error when connecting",
            str(ex),
            title="Kql response error",
        )

    @staticmethod
    def _raise_kql_engine_error(ex):
        ex_mssgs = [
            "An error was returned from Kqlmagic KqlEngine.",
            "This can occur if you tried to connect to a second workspace using a"
            + " different tenant ID - only a single tenant ID is supported in"
            + " one notebook.",
            "Other causes of this error could be an invalid format of your"
            + " connection string",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="kql connection error")

    @staticmethod
    def _raise_adal_error(ex):
        """Adal error - usually wrong tenant ID."""
        if ex.args[0] == "Unexpected polling state code_expired":
            raise MsticpyKqlConnectionError(
                "Authentication request was not completed.",
                title="authentication timed out",
            )

        err_response = getattr(ex, "error_response")
        if err_response and "error_description" in ex.error_response:
            ex_mssgs = ex.error_response["error_description"].split("\r\n")
        else:
            ex_mssgs = [f"Full error: {ex}"]
        raise MsticpyKqlConnectionError(
            *ex_mssgs, title="could not authenticate to tenant"
        )

    @staticmethod
    def _raise_authn_error(ex):
        """Raise an authentication error."""
        ex_mssgs = [
            "The authentication failed.",
            "Please check the credentials you are using and permissions on the workspace",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="authentication failed")

    @staticmethod
    def _raise_unknown_error(ex):
        """Raise an unknown exception."""
        raise MsticpyKqlConnectionError(
            "Another exception was returned by the service",
            *ex.args,
            f"Full exception:\n{str(ex)}",
            title="connection failed",
        )



"""TI Providers sub-package."""
from .alienvault_otx import OTX  # noqa:F401
from .ibm_xforce import XForce  # noqa:F401
from .virustotal import VirusTotal  # noqa:F401
from .azure_sent_byoti import AzSTI  # noqa:F401
from .open_page_rank import OPR  # noqa:F401
from .tor_exit_nodes import Tor  # noqa:F401
from .ti_provider_base import (  # noqa:F401
    TIProvider,
    preprocess_observable,
    LookupResult,
)
from .http_base import HttpProvider  # noqa:F401
from ...common.provider_settings import (  # noqa:F401
    ProviderSettings,
    get_provider_settings,
)
from ..._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Credential wrapper to expose ADAL and MSAL credentials."""
from azure.core.pipeline import PipelineContext, PipelineRequest
from azure.core.pipeline.policies import BearerTokenCredentialPolicy
from azure.core.pipeline.transport import HttpRequest
from azure.identity import DefaultAzureCredential
from msrest.authentication import BasicTokenAuthentication

from .._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"


# Class to extract v1 authentication token from DefaultAzureCredential object.
# Credit - https://gist.github.com/lmazuel/cc683d82ea1d7b40208de7c9fc8de59d
class CredentialWrapper(BasicTokenAuthentication):
    """Class for handling legacy auth conversion."""

    def __init__(
        self,
        credential=None,
        resource_id="https://management.azure.com/.default",
        **kwargs,
    ):
        """
        Wrap azure-identity credential to work with SDK that needs ADAL.

        Parameters
        ----------
        credential : [type], optional
            Any azure-identity credential, by default DefaultAzureCredential
        resource_id : str, optional
            The scope to use to get the token, by default
            "https://management.azure.com/.default"

        """
        super(CredentialWrapper, self).__init__(  # pylint: disable=super-with-arguments
            None
        )
        if credential is None:
            credential = DefaultAzureCredential()
        self._policy = BearerTokenCredentialPolicy(credential, resource_id, **kwargs)

    def set_token(self):
        """
        Ask the azure-core BearerTokenCredentialPolicy policy to get a token.

        Using the policy gives us for free the caching system of azure-core.
        We could make this code simpler by using private method, but by definition
        I can't assure they will be there forever, so mocking a fake call to the policy
        to extract the token, using 100% public API.

        """
        request = _make_request()
        self._policy.on_request(request)
        # Read Authorization, and get the second part after Bearer
        token = request.http_request.headers["Authorization"].split(" ", 1)[1]
        self.token = {"access_token": token}

    def signed_session(self, session=None):
        """Wrap signed session object."""
        self.set_token()
        return super(  # pylint: disable=super-with-arguments
            CredentialWrapper, self
        ).signed_session(session)


def _make_request():
    """Make mocked request to get token."""
    return PipelineRequest(
        HttpRequest("CredentialWrapper", "https://fakeurl"), PipelineContext(None)
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Deprecated path for common.utility.py."""
import warnings

# flake8: noqa: F403, F401
# pylint: disable=wildcard-import, unused-wildcard-import
from ..common.utility import *
from ..common.utility import md, md_warn


WARN_MSSG = (
    "This module has moved to msticpy.common.utility\n"
    + "Please change your import to reflect this new location."
)
warnings.warn(WARN_MSSG, category=DeprecationWarning)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for Log Analytics-related configuration."""

import os
import json
from typing import Dict, Any, Optional
from pathlib import Path
import warnings

from .exceptions import MsticpyUserConfigError
from .utility import export, is_valid_uuid
from . import pkg_config
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_NO_CONFIG_WARN = [
    "Could not find msticpyconfig.yaml or a config.json in the current directory.",
    "Using '{config_file}'.",
    "We recommend using an explicit msticpyconfig.yaml specified using the",
    "MSTICPYCONFIG environment variable. See",
    "https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html",
    "for more details.",
]

_NO_CONFIG_ERR = [
    "Could not find msticpyconfig.yaml or config.json."
    "The config.json file is created when you launch notebooks from",
    "Azure Sentinel. If you have copied the notebook to another location",
    "or folder you will need to copy this file."
    "Alternatively, we recommend using an explicit msticpyconfig.yaml",
    "and adding your Workspace and Tenant IDs to that file.",
]


@export
class WorkspaceConfig:
    """Workspace configuration class."""

    # Constants
    TENANT_ID = "{{cookiecutter.tenant_id}}"
    SUBSCRIPTION_ID = "{{cookiecutter.subscription_id}}"
    RESOURCE_GROUP = "{{cookiecutter.resource_group}}"
    WORKSPACE_ID = "{{cookiecutter.workspace_id}}"
    WORKSPACE_NAME = "{{cookiecutter.workspace_name}}"

    PKG_CONF_TENANT_KEY = "TenantId"
    PKG_CONF_WS_KEY = "WorkspaceId"

    CONF_WS_ID_KEY = "workspace_id"
    CONF_TENANT_ID_KEY = "tenant_id"
    CONF_SUB_ID_KEY = "subscription_id"
    CONF_RES_GROUP_KEY = "resource_group"
    CONF_WS_NAME_KEY = "workspace_name"

    def __init__(self, config_file: Optional[str] = None, workspace: str = None):
        """
        Load current Azure Notebooks configuration for Log Analytics.

        Parameters
        ----------
        config_file : Optional[str], optional
            path to a configuration file,
            Defaults to msticpyconfig.yaml if settings are configured there.
            If not, looks for a config.json in current folder
        workspace : str, optional
            Workspace name (where multiple workspaces are configured),
            by default the Default workspace will be used.

        """
        self._config: Dict[str, str] = {}
        if not config_file:
            self._read_pkg_config_values(workspace_name=workspace)
            if self.config_loaded:
                return
            if Path("./config.json").exists():
                config_file = "./config.json"
            else:
                config_file = self._search_for_file("**/config.json")
        self._config_file = config_file
        #
        config = self._read_config_values(config_file)
        if config:
            self._config.update(config)
        else:
            os.environ["MSTICPYCONFIG"] = config_file
            pkg_config.refresh_config()
            self._read_pkg_config_values(workspace_name=workspace)

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key in self._config:
            return self._config[key]
        raise KeyError

    def __setitem__(self, key: str, value: Any):
        """Allow property set using dictionary key syntax."""
        self._config[key] = value

    def __contains__(self, key: str):
        """Allow property in test."""
        # In operator overload
        return key == "Type" or key in self._config or key in self.__dict__

    def __repr__(self):
        """Return contents of current config."""
        return self._config.__repr__()

    @property
    def config_loaded(self) -> bool:
        """
        Return True if workspace id and tenant id have values.

        Returns
        -------
        bool
            True if configuration loaded.

        """
        ws_value = self._config.get(self.CONF_WS_ID_KEY, None)
        ten_value = self._config.get(self.CONF_TENANT_ID_KEY, None)
        return is_valid_uuid(ws_value) and is_valid_uuid(ten_value)  # type: ignore

    @property
    def code_connect_str(self) -> str:
        """
        Return the Log Analytics connection string for dev code auth.

        Returns
        -------
        str
            Connection string

        """
        ten_id = self._config.get(self.CONF_TENANT_ID_KEY, None)
        ws_id = self._config.get(self.CONF_WS_ID_KEY, None)
        if not ten_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_TENANT_ID_KEY} "
                + "could not be found."
            )
        if not ws_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_WS_ID_KEY} "
                + "could not be found."
            )
        return f"loganalytics://code().tenant('{ten_id}').workspace('{ws_id}')"

    @classmethod
    def _read_config_values(cls, file_path: str) -> Dict[str, str]:
        """Read configuration file."""
        try:
            with open(file_path) as json_file:
                if json_file:
                    return json.load(json_file)
        except json.JSONDecodeError:
            pass
        return {}

    @classmethod
    def list_workspaces(cls) -> Dict:
        """
        Return list of available workspaces.

        Returns
        -------
        Dict
            Dictionary of workspaces with workspace and tenantIds.

        """
        ws_settings = pkg_config.settings.get("AzureSentinel", {}).get("Workspaces")
        if not ws_settings:
            return {}
        return {
            ws_name: {
                cls.PKG_CONF_WS_KEY: ws.get(cls.PKG_CONF_WS_KEY),
                cls.PKG_CONF_TENANT_KEY: ws.get(cls.PKG_CONF_TENANT_KEY),
            }
            for ws_name, ws in ws_settings.items()
        }

    def _read_pkg_config_values(self, workspace_name: str = None):
        as_settings = pkg_config.settings.get("AzureSentinel")
        if not as_settings:
            return {}
        ws_settings = as_settings.get("Workspaces")  # type: ignore
        if not ws_settings:
            return {}
        if workspace_name and workspace_name in ws_settings:
            selected_workspace = ws_settings[workspace_name]
        elif "Default" in ws_settings:
            selected_workspace = ws_settings["Default"]
        else:
            return {}
        if (
            selected_workspace
            and self.PKG_CONF_WS_KEY in selected_workspace
            and self.PKG_CONF_TENANT_KEY in selected_workspace
        ):
            self._config[self.CONF_WS_ID_KEY] = selected_workspace.get(
                self.PKG_CONF_WS_KEY
            )
            self._config[self.CONF_TENANT_ID_KEY] = selected_workspace.get(
                self.PKG_CONF_TENANT_KEY
            )
        return {}

    def _search_for_file(self, pattern: str) -> str:
        config_file = None
        searched_configs = list(Path("..").glob(pattern))
        if not searched_configs:
            raise MsticpyUserConfigError(
                *_NO_CONFIG_ERR, title="Workspace configuration missing."
            )
        for found_file in searched_configs:
            test_content = self._read_config_values(str(found_file))
            if "workspace_id" in test_content:
                config_file = found_file
                break
        if config_file is None:
            raise MsticpyUserConfigError(
                *_NO_CONFIG_ERR, title="Workspace configuration missing."
            )
        # Warn that we're using a "found" file
        warnings.warn("\n".join(_NO_CONFIG_WARN).format(config_file=config_file))
        return str(config_file)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""LogAnayltics Uploader class."""
from typing import Any
import json
import datetime
import sys
import hashlib
import hmac
import base64
import re
from pathlib import Path

import requests
from tqdm.notebook import tqdm
import pandas as pd

from ...common.exceptions import MsticpyConnectionError

from .uploader_base import UploaderBase
from ..._version import VERSION

# Credits
# https://docs.microsoft.com/en-us/azure/azure-monitor/platform/data-collector-api#python-3-sample
# https://github.com/Cyb3rWard0g/azure-loganalytics-api-clients/blob/master/python/ala-python-data-producer.py


__version__ = VERSION
__author__ = "Pete Bryan"


class LAUploader(UploaderBase):
    """Uploader class for LogAnalytics."""

    def __init__(self, workspace: str, workspace_secret: str, **kwargs):
        """Initialize a LogAnalytics Uploader instance."""
        super().__init__()
        self._kwargs = kwargs
        self.workspace = workspace
        self.workspace_secret = workspace_secret
        self._debug = kwargs.get("debug", False)
        self.ops_loc = kwargs.get("opsinsight_loc", ".ods.opinsights.azure.com")

    def _build_signature(
        self,
        date: str,
        content_length: int,
        method: str,
        content_type: str,
        resource: str,
    ) -> str:
        """
        Build authentication authentication string to pass to LA API.

        Parameters
        ----------
        date : str
            datetime of authencation session.
        content_length : int
            lenght of content to be passed to the api.
        method : str
            HTTP method being used.
        content_type : str
            Type of content being passed to the API.
        resource : str
            The API endpoint being targetted.

        Returns
        -------
        str
            The encoded authorization string.

        """
        x_headers = "x-ms-date:" + date
        string_to_hash = "\n".join(
            [method, str(content_length), content_type, x_headers, resource]
        )
        bytes_to_hash = bytes(string_to_hash, encoding="utf-8")
        decoded_key = base64.b64decode(self.workspace_secret)
        encoded_hash = base64.b64encode(
            hmac.new(decoded_key, bytes_to_hash, digestmod=hashlib.sha256).digest()
        ).decode()
        authorization = f"SharedKey {self.workspace}:{encoded_hash}"
        return authorization

    def _post_data(self, body: str, table_name: str):
        """
        Write data to Log Analytics Workspace.

        Parameters
        ----------
        body : str
            The JSON formatted data to write to Log Analytics.
        table_name : str
            The name of the custom table to write the data to.

        Raises
        ------
        MsticpyConnectionError
            Raised when response code indicates failure.

        """
        table_name = re.sub("[^A-Za-z0-9_]+", "", table_name)

        resource = "/api/logs"
        content_type = "application/json"
        rfc1123date = datetime.datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        content_length = len(body)
        signature = self._build_signature(
            rfc1123date, content_length, "POST", content_type, resource
        )
        uri = (
            "https://"
            + self.workspace
            + self.ops_loc
            + resource
            + "?api-version=2016-04-01"
        )
        headers = {
            "content-type": content_type,
            "Authorization": signature,
            "Log-Type": table_name,
            "x-ms-date": rfc1123date,
        }
        try:
            response = requests.post(uri, data=body, headers=headers)
        except requests.ConnectionError as req_err:
            raise MsticpyConnectionError(
                "Unable to connect to workspace, ensure your Workspace ID is correct.",
                title="Unable to connect to Workspace",
            ) from req_err
        if self._debug is True:
            print(f"Upload response code: {response.status_code}")
        if response.status_code < 200 or response.status_code > 299:
            raise MsticpyConnectionError(
                f"""LogAnalytics data upload failed with code {response.status_code}.
                Check Workspace ID and key""",
                title="Data Upload Failed",
            )

    def upload_df(self, data: pd.DataFrame, table_name: Any, **kwargs):
        """
        Upload a pandas DataFrame to Log Analytics.

        Parameters
        ----------
        data : pd.DataFrame
            Pandas DataFrame to upload.
        table_name : str
            Custom table name to upload the data to.

        """
        events = []
        for row in data.iterrows():
            events.append(row[1].astype(str).to_dict())
            # Due to 30MB limit if data is larger than 25Mb upload that chunk then continue
            if sys.getsizeof(json.dumps(events)) > 26214400:
                if self._debug is True:
                    print("Data larger than 25MB spliting data requests.")
                body = json.dumps(events)
                self._post_data(body, table_name)
                events = []

        if events:
            body = json.dumps(events)
            self._post_data(body, table_name)

        if self._debug:
            print(f"Upload to {table_name} complete")

    def upload_file(
        self, file_path: str, table_name: str = None, delim: str = ",", **kwargs
    ):
        """
        Upload a seperated value file to Log Analytics.

        Parameters
        ----------
        file_path : str
            Path to file to upload.
        table_name : str
            Table name to upload data to.
        delim : str, optional
            Value seperator used by the file, by default ","

        """
        path = Path(file_path)
        data = pd.read_csv(path, delimiter=delim)
        if not table_name:
            table_name = path.stem
        self.upload_df(data, table_name)

    def upload_folder(
        self, folder_path: str, table_name: str = None, delim: str = ",", **kwargs
    ):
        """
        Upload all files in a folder to Log Analytics.

        Parameters
        ----------
        folder_path : str
            Folder path to upload.
        table_name : str, optional
            Table name to upload all files to, by default None.
        delim : str, optional
            Sperator used in files in target folder, by default ",".

        """
        glob_pat = kwargs.get("glob", "*")
        ext = glob_pat
        t_name = bool(table_name)
        input_files = Path(folder_path).glob(ext)
        # pylint: disable=unnecessary-comprehension
        input_files = [path for path in input_files]  # type: ignore
        # pylint: enable=unnecessary-comprehension
        progress = tqdm(total=len(list(input_files)), desc="Files", position=0)
        for path in input_files:
            data = pd.read_csv(path, delimiter=delim)
            if t_name is False:
                table_name = path.stem
            self.upload_df(data, table_name)
            progress.update(1)
        progress.close()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Wrapper module for Model class for modelling sessions.

In particular, this module is for both modelling and visualising your session data.
"""

import pandas as pd

from ...nbtools import timeline
from .model import Model
from ...common.exceptions import MsticpyException


def score_sessions(
    data: pd.DataFrame, session_column: str, window_length: int
) -> pd.DataFrame:
    """
    Model sessions using a sliding window approach within a markov model.

    Parameters
    ----------
    data: pd.DataFrame
        Dataframe which contains at least a column for sessions
    session_column: str
        name of the column which contains the sessions
        The values in the session column should take one of the following formats:

        1) ['Set-User', 'Set-Mailbox']
        2) [Cmd(name='Set-User', params={'Identity', 'Force'}),
           Cmd(name='Set-Mailbox', params={'Identity', 'AuditEnabled'})]
        3) [Cmd(name='Set-User',
           params={'Identity': 'blahblah', 'Force': 'true'}),
           Cmd(name='Set-Mailbox',
           params={'Identity': 'blahblah', 'AuditEnabled': 'false'})]

        The Cmd datatype can be accessed from
        anomalous_sequence.utils.data_structures.Cmd
    window_length: int
        length of the sliding window to use when computing the likelihood
        metrics for each session.
        This should be set to an integer >= 2. Note that sessions which have
        fewer commands than the chosen window_length + 1 will end up with a
        np.nan score. (The + 1 is because we append a dummy `end_token` to each
        session before starting the sliding window, so a session of length 2,
        would be treated as length 3)

    Returns
    -------
    input dataframe with two additional columns appended.

    """
    if not isinstance(data, pd.DataFrame):
        raise MsticpyException("`data` should be a pandas dataframe")
    if session_column not in data.columns:
        raise MsticpyException(
            '"{}" should be a column in the `data`'.format(session_column)
        )

    sessions_df = data.copy()
    sessions = sessions_df[session_column].values.tolist()

    model = Model(sessions=sessions)
    model.train()
    model.compute_rarest_windows(
        window_len=window_length, use_geo_mean=False, use_start_end_tokens=True
    )

    sessions_df[
        "rarest_window{}_likelihood".format(window_length)
    ] = model.rare_window_likelihoods[window_length]
    sessions_df["rarest_window{}".format(window_length)] = model.rare_windows[
        window_length
    ]

    return sessions_df


# pylint: disable=too-many-arguments
def visualise_scored_sessions(
    data_with_scores: pd.DataFrame,
    time_column: str,
    score_column: str,
    window_column: str,
    score_upper_bound: float = None,
    source_columns: list = None,
):
    """
    Visualise the scored sessions on an interactive timeline.

    Parameters
    ----------
    data_with_scores: pd.DataFrame
        Dataframe which contains at least columns for time,
        session score, window representing the session
    time_column: str
        name of the column which contains a timestamp
    score_column: str
        name of the column which contains a numerical score for each
        of the sessions
    window_column: str
        name of the column which contains a representation of each of the sessions.
        This representation will appear in the tooltips in the figure.
        For example, it could be the rarest window of the session,
        or the full session etc.
    score_upper_bound: float, optional
        an optional upper bound on the score for the visualisation figure.
        This can help to zoom in on the more anomalous sessions
    source_columns: list, optional
        an optional list of source columns to include in the tooltips
        in the visualisation.
        Note, the content of each of these columns should be json serializable
        in order to be compatible with the figure

    Returns
    -------
    figure

    """
    scored_sessions = data_with_scores.copy()  # so we don't affect input dataframe
    scored_sessions[window_column] = scored_sessions[window_column].apply(
        lambda x: [str(cmd) for cmd in x]
    )  # so it
    # becomes json serializable for the figure
    scored_sessions["time_col"] = scored_sessions[time_column].astype(
        str
    )  # so it appears nicely in the figure
    # tooltips

    if score_upper_bound is None:
        score_upper_bound = scored_sessions[score_column].max()

    if source_columns is None:
        source_columns = []
    source_columns += [score_column, window_column, "time_col"]
    source_columns = list(set(source_columns))

    timeline.display_timeline_values(
        data=scored_sessions.loc[scored_sessions[score_column] <= score_upper_bound],
        y=score_column,
        time_column=time_column,
        source_columns=source_columns,
        kind="circle",
    )


# pylint: disable=too-many-arguments
def score_and_visualise_sessions(
    data: pd.DataFrame,
    session_column: str,
    window_length: int,
    time_column: str,
    likelihood_upper_bound: float = None,
    source_columns: list = None,
):
    """
    Model sessions and then produces an interactive timeline visualisation plot.

    In particular, the sessions are modelled using a sliding window approach
    within a markov model. The visualisation plot has time on the x-axis and
    the modelled session likelihood metric on the y-axis.

    Parameters
    ----------
    data: pd.DataFrame
        Dataframe which contains at least columns for time and sessions
    session_column: str
        name of the column which contains the sessions
        The values in the session column should take one of the following formats:

        1) ['Set-User', 'Set-Mailbox']
        2) [Cmd(name='Set-User', params={'Identity', 'Force'}),
           Cmd(name='Set-Mailbox', params={'Identity', 'AuditEnabled'})]
        3) [Cmd(name='Set-User',
           params={'Identity': 'blahblah', 'Force': 'true'}),
           Cmd(name='Set-Mailbox',
           params={'Identity': 'blahblah', 'AuditEnabled': 'false'})]

        The Cmd datatype can be accessed from
        seqeunce.utils.data_structures.Cmd
    window_length: int
        length of the sliding window to use when computing the
        likelihood metrics for each session.

        This should be set to an integer >= 2.
        Note that sessions which have fewer commands than the chosen
        window_length + 1 will not appear in the visualisation. (The + 1 is
        because we append a dummy `end_token` to each session before starting
        the sliding window, so a session of length 2, would be treated as length
        3)
    time_column: str
        name of the column which contains a timestamp
    likelihood_upper_bound: float, optional
        an optional upper bound on the likelihood metrics for the visualisation
        plot. This can help to zoom in on the more anomalous sessions
    source_columns: list, optional
        An optional list of source columns to include in the tooltips
        in the visualisation.
        Note, the content of each of these columns should be json
        serializable in order to be compatible with the figure

    Returns
    -------
    figure

    """
    scored_sessions = score_sessions(
        data=data, session_column=session_column, window_length=window_length
    )
    score_column = "rarest_window{}_likelihood".format(window_length)
    window_column = "rarest_window{}".format(window_length)

    visualise_scored_sessions(
        data_with_scores=scored_sessions,
        time_column=time_column,
        score_column=score_column,
        window_column=window_column,
        score_upper_bound=likelihood_upper_bound,
        source_columns=source_columns,
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Open Page Rank Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from json import JSONDecodeError
from typing import Any, Tuple, Union, Iterable, Dict, List

import attr
import pandas as pd

from .ti_provider_base import LookupResult, TISeverity, generate_items, TILookupStatus
from .http_base import HttpProvider, IoCLookupParams
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class OPR(HttpProvider):
    """Open PageRank Lookup."""

    _BASE_URL = "https://openpagerank.com"

    _IOC_QUERIES = {
        "dns": IoCLookupParams(
            path="/api/v1.0/getPageRank",
            params={"domains[0]": "{observable}"},
            headers={"API-OPR": "{API_KEY}"},
        )
    }

    _REQUIRED_PARAMS = ["API_KEY"]

    def __init__(self, **kwargs):
        """Initialize a new instance of the class."""
        super().__init__(**kwargs)

        self._provider_name = self.__class__.__name__
        print(
            "Using Open PageRank.",
            "See https://www.domcop.com/openpagerank/what-is-openpagerank",
        )

    # pylint: disable=duplicate-code
    def lookup_iocs(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        obs_col: str = None,
        ioc_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup collection of IoC observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        obs_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        kwargs.get("provider_name", self.__class__.__name__)

        domain_list = set()
        bad_requests: List[pd.Series] = []
        for ioc, ioc_type in generate_items(data, obs_col, ioc_type_col):
            if not ioc:
                continue
            result = self._check_ioc_type(
                ioc=ioc, ioc_type=ioc_type, query_subtype=query_type
            )

            if result.status == TILookupStatus.ok.value:
                domain_list.add(result.ioc)
            else:
                bad_requests.append(pd.Series(attr.asdict(result)))

        results: List[pd.Series] = []
        if not domain_list:
            return pd.DataFrame(columns=LookupResult.column_map())
        for item_result in self._lookup_bulk_request(domain_list):  # type: ignore
            results.append(pd.Series(attr.asdict(item_result)))

        all_results = results + bad_requests
        return pd.DataFrame(data=all_results).rename(columns=LookupResult.column_map())

    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(response.raw_result, dict):
            return False, TISeverity.information, "Not found."

        severity = TISeverity.information
        if "response" in response.raw_result:
            dom_records = response.raw_result["response"]
            dom_record = dom_records[0]
            return self._parse_one_record(dom_record)
        return True, severity, {}

    def _parse_multi_results(self, response: LookupResult) -> Iterable[LookupResult]:
        """Parse details of batch response."""
        if not isinstance(response.raw_result, dict):
            new_result = LookupResult(**attr.asdict(response))
            new_result.result = False
            new_result.set_severity(TISeverity.information)
            new_result.details = "Not found."
            yield new_result
        elif "response" in response.raw_result:
            dom_records = response.raw_result["response"]
            for dom_record in dom_records:
                result, sev, details = self._parse_one_record(dom_record)
                domain_name = dom_record["domain"]
                new_result = LookupResult(ioc=domain_name, ioc_type="dns")
                new_result.ioc = domain_name
                new_result.provider = self._provider_name
                new_result.result = result
                new_result.set_severity(sev)
                new_result.details = details
                new_result.raw_result = dom_record
                new_result.reference = f"{response.reference}?domains[0]={domain_name}"
                yield new_result

    @staticmethod
    def _parse_one_record(dom_record: dict):
        record_status = dom_record.get("status_code", 404)
        severity = TISeverity.information
        if record_status == 200:
            return (
                True,
                severity,
                {
                    "rank": dom_record.get("rank", "0"),
                    "page_rank": dom_record.get("page_rank_decimal", 0),
                    "error": dom_record.get("error", ""),
                },
            )
        if record_status == 404:
            return (
                True,
                TISeverity.warning,
                {
                    "rank": dom_record.get("rank", "0"),
                    "error": dom_record.get("error", ""),
                },
            )
        return False, TISeverity.information, {}

    def _lookup_bulk_request(self, ioc_list: Iterable[str]) -> Iterable[LookupResult]:
        ioc_list = list(ioc_list)
        batch_size = 100

        l_len = len(ioc_list)
        for step in range(0, l_len, batch_size):
            batch_list = ioc_list[step : (step + batch_size)]  # noqa: E203
            for result in self._lookup_batch(batch_list):
                yield result

    # pylint: disable=duplicate-code
    def _lookup_batch(self, ioc_list: list) -> Iterable[LookupResult]:
        # build the query string manually - of the form domains[N]=domN&domains[N+1]...
        qry_elements = []
        for idx, dom in zip(range(0, len(ioc_list)), ioc_list):
            qry_elements.append(f"domains[{idx}]={dom}")
        qry_str = "&".join(qry_elements)
        path = self._IOC_QUERIES["dns"].path
        req_url = f"{self._BASE_URL}{path}?{qry_str}"

        try:
            _, req_params = self._substitute_parms("dummy", "dns", None)
            response = self._requests_session.get(
                url=req_url, headers=req_params["headers"]
            )
            result = LookupResult(ioc=",".join(ioc_list), ioc_type="dns")
            if response.status_code == 200:
                result.status = TILookupStatus.ok.value
                result.reference = self._BASE_URL + path
                result.raw_result = response.json()
                for single_result in self._parse_multi_results(result):
                    yield single_result
            else:
                result.raw_result = str(response)
                result.result = False
                result.reference = req_url
                result.status = response.status_code
                result.details = "No response from provider."
                yield result
        except (
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            self._err_to_results(result, err)
            if not isinstance(err, LookupError):
                result.reference = req_url
            yield result

    # pylint: enable=duplicate-code



"""Data sub-package."""
# flake8: noqa: F403
from .data_providers import QueryProvider
from .azure_data import AzureData
from .query_defns import DataEnvironment, DataFamily

from .._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure authentication handling."""
import os
from typing import List

from azure.common.exceptions import CloudError
from azure.mgmt.subscription import SubscriptionClient

from .._version import VERSION
from .azure_auth_core import AzCredentials, az_connect_core
from .provider_settings import get_provider_settings

__version__ = VERSION
__author__ = "Pete Bryan"


def az_connect(
    auth_methods: List = None,
    silent: bool = False,
) -> AzCredentials:
    """
    Connect to Azure SDK/API.

    Parameters
    ----------
    auth_methods : List, optional
        List of authentication methods to try
        Possible options are:
        - "env" - to get authentication details from environment varibales
        - "cli" - to use Azure CLI authentication details
        - "msi" - to user Managed Service Indenity details
        - "interactive" - to prompt for interactive login
        Default is ["env", "cli", "msi", "interactive"]
    silent : bool, optional
        Set True to hide all output during connection, by default False

    Returns
    -------
    AzCredentials
        Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials


    Raises
    ------
    CloudError
        If chained token credential creation fails.

    """
    # If using env options try to load from msticpy
    if auth_methods is None or "env" in auth_methods:
        try:
            data_provs = get_provider_settings(config_section="DataProviders")
            az_cli_config = data_provs.get("AzureCLI")
            config_items = az_cli_config.args  # type: ignore
            os.environ["AZURE_CLIENT_ID"] = config_items["clientId"]
            os.environ["AZURE_TENANT_ID"] = config_items["tenantId"]
            os.environ["AZURE_CLIENT_SECRET"] = config_items["clientSecret"]
        except KeyError:
            pass
    credentials = az_connect_core(auth_methods=auth_methods, silent=silent)
    sub_client = SubscriptionClient(credentials.modern)  # type: ignore
    if not sub_client:
        raise CloudError("Could not create a Subscription client.")

    return credentials


def az_user_connect(silent: bool = False) -> AzCredentials:
    """
    Authenticate to the SDK using user based authenticaiton methods, Azure CLI or interactive logon.

    Parameters
    ----------
    silent : bool, optional
        Whether you want the auth process to display any output, by default False

    Returns
    -------
    AzCredentials

    """
    credentials = az_connect_core(auth_methods=["cli", "interactive"], silent=silent)
    return credentials



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""

Functions to support investigation of a domain or url.

Includes functions to conduct common investigation steps when dealing
with a domain or url, such as getting a screenshot or validating the TLD.

"""
import json
import ssl
import sys
import time
import warnings
from pathlib import Path
from typing import Set, Tuple, Optional
from urllib.error import HTTPError, URLError

import cryptography as crypto
import dns.resolver
import pandas as pd
import pkg_resources
import requests
import tldextract
from IPython import display
from ipywidgets import IntProgress

from .._version import VERSION
from ..common import pkg_config as config
from ..common.exceptions import MsticpyUserConfigError
from ..common.utility import export

__version__ = VERSION
__author__ = "Pete Bryan"


@export
def screenshot(url: str, api_key: str = None) -> requests.models.Response:
    """
    Get a screenshot of a url with Browshot.

    Parameters
    ----------
    url : str
        The url a screenshot is wanted for.
    api_key : str (optional)
        Browshot API key. If not set msticpyconfig checked for this.

    Returns
    -------
    image_data: requests.models.Response
        The final screenshot request response data.

    """
    # Get Browshot API key from kwargs or config
    if api_key is not None:
        bs_api_key: Optional[str] = api_key
    else:
        bs_conf = config.settings.get("DataProviders", {}).get(
            "Browshot"
        ) or config.settings.get("Browshot")
        bs_api_key = None
        if bs_conf is not None:
            bs_api_key = bs_conf.get("Args", {}).get("AuthKey")  # type: ignore

    if bs_api_key is None:
        raise MsticpyUserConfigError(
            "No configuration found for Browshot",
            "Please add a section to msticpyconfig.yaml:",
            "DataProviders:",
            "  Browshot:",
            "    Args:",
            "      AuthKey: {your_auth_key}",
            title="Browshot configuration not found",
            browshot_uri=("Get an API key for Browshot", "https://api.browshot.com/"),
        )

    # Request screenshot from Browshot and get request ID
    id_string = f"https://api.browshot.com/api/v1/screenshot/create?url={url}/&instance_id=26&size=screen&cache=0&key={bs_api_key}"  # pylint: disable=line-too-long
    id_data = requests.get(id_string)
    bs_id = json.loads(id_data.content)["id"]
    status_string = (
        f"https://api.browshot.com/api/v1/screenshot/info?id={bs_id}&key={bs_api_key}"
    )
    image_string = f"https://api.browshot.com/api/v1/screenshot/thumbnail?id={bs_id}&zoom=50&key={bs_api_key}"  # pylint: disable=line-too-long
    # Wait until the screenshot is ready and keep user updated with progress
    print("Getting screenshot")
    progress = IntProgress(min=0, max=40)
    display.display(progress)
    ready = False
    while ready is False:
        progress.value += 1
        status_data = requests.get(status_string)
        status = json.loads(status_data.content)["status"]
        if status == "finished":
            ready = True
        else:
            time.sleep(0.05)
    progress.value = 40

    # Once ready get the screenshot
    image_data = requests.get(image_string)

    if image_data.status_code != 200:
        print(
            "There was a problem with the request, please check the status code for details"
        )

    return image_data


@export
class DomainValidator:
    """Assess a domain's validity."""

    _tld_index: Set[str] = set()
    _ssl_abuse_list: pd.DataFrame = pd.DataFrame()

    @classmethod
    def _check_and_load_tlds(cls):
        """Pull IANA TLD list and save to internal attribute."""
        if not cls._tld_index:
            cls._tld_index: Set[str] = cls._get_tlds()

    @classmethod
    def _check_and_load_abuselist(cls):
        """Pull IANA TLD list and save to internal attribute."""
        if cls._ssl_abuse_list is None or cls._ssl_abuse_list.empty:
            cls._ssl_abuse_list: pd.DataFrame = cls._get_ssl_abuselist()

    @property
    def tld_index(self) -> Set[str]:
        """
        Return the class TLD index.

        Returns
        -------
        Set[str]
            The current TLD index

        """
        self._check_and_load_tlds()
        return self._tld_index

    @property
    def ssl_abuse_list(self) -> pd.DataFrame:
        """
        Return the class SSL Blacklist.

        Returns
        -------
        pd.DataFrame
            SSL Blacklist

        """
        self._check_and_load_abuselist()
        return self._ssl_abuse_list

    def validate_tld(self, url_domain: str) -> bool:
        """
        Validate if a domain's TLD appears in the IANA tld list.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid public TLD, False if not.

        """
        if not self.tld_index:
            warnings.warn(
                f"No top-level domain list available - {url_domain}"
                + " could not be verified. ",
                RuntimeWarning,
            )
            return True
        _, _, tld = tldextract.extract(url_domain.lower())
        if "." in tld:
            ttld = tld.split(".")[1].upper()
        else:
            ttld = tld.upper()
        return ttld in self.tld_index

    def is_resolvable(self, url_domain: str) -> bool:  # pylint: disable=no-self-use
        """
        Validate if a domain or URL be be resolved to an IP address.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid resolvable, False if not.

        """
        try:
            dns.resolver.query(url_domain, "A")
            result = True
        except Exception:  # pylint: disable=broad-except
            result = False

        return result

    def in_abuse_list(self, url_domain: str) -> Tuple:
        """
        Validate if a domain or URL's SSL cert the abuse.ch SSL Abuse List.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid in the list, False if not.

        """
        try:
            cert = ssl.get_server_certificate((url_domain, 443))
            backend = crypto.hazmat.backends.default_backend()  # type: ignore
            x509 = crypto.x509.load_pem_x509_certificate(  # type: ignore
                cert.encode("ascii"), backend
            )
            cert_sha1 = x509.fingerprint(
                crypto.hazmat.primitives.hashes.SHA1()  # type: ignore # nosec
            )
            result = bool(
                self.ssl_abuse_list["SHA1"]
                .str.contains(cert_sha1.hex())
                .any()  # type: ignore
            )
        except Exception:  # pylint: disable=broad-except
            result = False
            x509 = None

        return result, x509

    @classmethod
    def _get_tlds(cls) -> Set[str]:
        """
        Return IANA Top Level Domains.

        Returns
        -------
        Set[str]
            Set of top level domains.

        """
        try:
            tld_list = "https://data.iana.org/TLD/tlds-alpha-by-domain.txt"
            temp_df = pd.read_csv(tld_list, skiprows=1, names=["TLD"])
            return set(temp_df["TLD"].dropna())
        except (HTTPError, URLError):
            pass
        except Exception as err:  # pylint: disable=broad-except
            warnings.warn(
                "Exception detected trying to retrieve IANA top-level domain list."
                + "Falling back to builtin seed list. "
                + f"{err.args}",
                RuntimeWarning,
            )
        # pylint: enable=broad-except
        # if we failed to get the list try to read from a seed file
        return cls._read_tld_seed_file()

    @classmethod
    def _read_tld_seed_file(cls) -> Set[str]:
        """Read TLD seed list from seed file."""
        seed_file = "tld_seed.txt"
        conf_file = pkg_resources.resource_filename(__name__, seed_file)

        if not Path(conf_file).is_file():
            # if all else fails we try to find the package default config somewhere
            # in the package tree - we use the first one we find
            pkg_paths = sys.modules["msticpy"]
            if pkg_paths:
                conf_file = str(
                    next(Path(pkg_paths.__path__[0]).glob(seed_file))  # type: ignore
                )

        if conf_file:
            with open(conf_file, "r") as file_handle:
                tld_txt = file_handle.read()
                tld_set = set(tld_txt.split("\n"))
            return tld_set
        return set()

    @classmethod
    def _write_tld_seed_file(cls):
        """Write existing TLD list to a text file."""
        if cls._tld_index:
            seed_file = "tld_seed.txt"
            with open(seed_file, "w") as file_handle:
                file_handle.write("\n".join(sorted(cls.tld_index)))

    @classmethod
    def _get_ssl_abuselist(cls) -> pd.DataFrame:
        """Download and load abuse.ch SSL Abuse List."""
        try:
            ssl_ab_list = pd.read_csv(
                "https://sslbl.abuse.ch/blacklist/sslblacklist.csv", skiprows=8
            )
        except (ConnectionError, HTTPError, URLError):
            ssl_ab_list = pd.DataFrame({"SHA1": []})

        return ssl_ab_list



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Miscellaneous helper methods for Jupyter Notebooks."""
import builtins
import difflib
import os
import re
import subprocess  # nosec
import sys
import uuid
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pkg_resources
from deprecated.sphinx import deprecated
from IPython import get_ipython
from IPython.core.display import HTML, display
from tqdm import tqdm, tqdm_notebook

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def export(func: Callable):
    """Decorate function or class to export to __all__."""
    mod = sys.modules[func.__module__]
    if hasattr(mod, "__all__"):
        all_list = getattr(mod, "__all__")
        all_list.append(func.__name__)
    else:
        all_list = [func.__name__]
        setattr(mod, "__all__", all_list)
    return func


@export
def string_empty(string: str) -> bool:
    """Return True if the input string is None or whitespace."""
    return (string is None) or not (string and string.strip())


@export
def is_not_empty(test_object: Any) -> bool:
    """Return True if the test_object is not None or empty."""
    if test_object:
        if isinstance(test_object, str):
            if test_object.strip():
                # test_object is not None AND myString is not empty or blank
                return True
            return False
        return True
    return False


# Toggle Code Cell Contents
_TOGGLE_CODE_STR = """
<form action="javascript:code_toggle()">
    <input type="submit" id="toggleButton" value="Show/Hide Code">
</form>
"""

_TOGGLE_CODE_PREPARE_STR = """
    <script>
    function code_toggle() {
        if ($('div.cell.code_cell.rendered.selected div.input').css('display')!='none'){
            $('div.cell.code_cell.rendered.selected div.input').hide();
        } else {
            $('div.cell.code_cell.rendered.selected div.input').show();
        }
    }
    </script>

"""


@export
def enable_toggle_code():
    """Load JS Function to enable code toggle button."""
    display(HTML(_TOGGLE_CODE_PREPARE_STR))


@export
def toggle_code():
    """Display a toggle button to hide/reveal code cell."""
    display(HTML(_TOGGLE_CODE_STR))


# String escapes
@export
def escape_windows_path(str_path: str) -> str:
    """Escape backslash characters in a string."""
    if is_not_empty(str_path):
        return str_path.replace("\\", "\\\\")
    return str_path


@export
def unescape_windows_path(str_path: str) -> str:
    """Remove escaping from backslash characters in a string."""
    if is_not_empty(str_path):
        return str_path.replace("\\\\", "\\")
    return str_path


@deprecated(reason="Inline Javascript no longer supported", version="0.3.2")
@export
def get_nb_query_param(nb_url_search: str, param: str) -> Optional[str]:
    """
    Get a url query parameter from the search string.

    Parameters
    ----------
    nb_url_search: str
        The URL search string
    param: str
        The parameter name to search for

    Returns
    -------
    Optional[str]
        value of the query string parameter or None if not found.

    """
    qs_regex = r"[\\?&]{param}=(?P<val>[^&#]*)".format(param=param)
    query_string_match = re.search(qs_regex, nb_url_search)
    if query_string_match:
        return query_string_match["val"]
    return None


@deprecated(reason="Inline Javascript no longer supported", version="0.3.2")
@export
def get_nb_query_params(nb_url_search: str) -> dict:
    """
    Get the url query parameters from the search string.

    Parameters
    ----------
    nb_url_search : str
        The URL search string

    Returns
    -------
    dict
        dictionary of the query string parameters.

    """
    nb_params = {}
    query_string_match = re.search(r"\?(?P<qs>[^#]+)#?", nb_url_search)
    if query_string_match:
        for param in query_string_match["qs"].split("&"):
            if "=" in param:
                nb_params[param.split("=")[0]] = param.split("=")[1]
    return nb_params


@deprecated(reason="Inline Javascript no longer supported", version="0.3.2")
@export
def get_notebook_query_string():
    """Execute javascript to publish notebook query string as python variable."""
    HTML(
        """
    <script type="text/javascript">
        IPython.notebook.kernel.execute(
            "nb_query_string='".concat(window.location.search).concat("'"));
    </script>
    """
    )


@export
def check_py_version(min_ver: Tuple = (3, 6)):
    """
    Check that the current python version is not less than `min_ver`.

    Parameters
    ----------
    min_ver : Tuple, optional
        Minimum required version, by default (3,6)

    """
    if isinstance(min_ver, (float, str)):
        min_ver_list = str(min_ver).split(".")
        min_ver = (int(min_ver_list[0]), int(min_ver_list[1]))
    if sys.version_info < min_ver:
        print("Check the Kernel->Change Kernel menu and ensure that Python 3.6")
        print("or later is selected as the active kernel.")
        raise SystemExit(
            "Python %s.%s or later is required.\n" % (min_ver[0], min_ver[1])
        )


@export
def resolve_pkg_path(part_path: str):
    """
    Resolve a path relative to the package.

    Parameters
    ----------
    part_path : str
        Absolute or relative path to resolve.

    """
    if Path(part_path).is_absolute():
        return part_path

    resolved_path = str(Path(__file__).resolve().parent.parent.joinpath(part_path))
    if Path(resolved_path).exists():
        return str(resolved_path)

    searched_paths = list(
        Path(__file__).resolve().parent.parent.glob(str(Path("**").joinpath(part_path)))
    )
    if not searched_paths or len(searched_paths) > 1:
        warnings.warn(f"No path or ambiguous match for {part_path} not found")
        return None
    return str(searched_paths[0])


# pylint: disable=not-an-iterable, too-many-branches
@export  # noqa: MC0001
def check_and_install_missing_packages(
    required_packages: List[str],
    force_notebook: bool = False,
    user: bool = True,
    upgrade: bool = False,
) -> bool:
    """
    Check and install missing packages from provided list of packages.

    Parameters
    ----------
    required_packages : List[str]
        List of packages to check and install in a current environment
        Note you can add package version constraints by appending them to
        the package name, e.g. `pandas>=1.01`
    force_notebook : bool, optional
        Boolean value to force notebook version of progress bar,
        by default False (autodetect)
    user : bool, optional
        Boolean value to toggle user flag while installing pip packages,
        by default True
    upgrade: bool, option
        If true supply `--upgrade` flag to pip to install the latest
        version (applies to all package in `required_packages`)

    Returns
    -------
    bool :
        True if successful, else False

    """
    missing_packages = []
    # Check package requirements against installed set
    for req in required_packages:
        pkg_req = pkg_resources.Requirement.parse(req)
        try:
            found_pkg = pkg_resources.working_set.find(pkg_req)
        except pkg_resources.VersionConflict:
            found_pkg = None
        if found_pkg is None:
            missing_packages.append(req)

    if not missing_packages:
        print("All packages are already installed")
        return True

    print("Missing packages to be installed: ", *missing_packages, sep=" ")
    if is_ipython() or force_notebook:
        pkgbar = tqdm_notebook(missing_packages, desc="Installing...", unit="bytes")
    else:
        pkgbar = tqdm(missing_packages, desc="Installing...", unit="bytes")

    pkg_command = ["pip", "install"]
    if user:
        pkg_command.append("--user")
    if upgrade:
        pkg_command.append("--upgrade")
    pkg_success = True
    for package in pkgbar:
        try:
            subprocess.run(  # nosec
                pkg_command + [package],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as proc_err:
            print(f"An Error has occured while installing {package}.")
            print(f"Output: {str(proc_err.stdout)}")
            print(f"Errs: {str(proc_err.stderr)}")
            pkg_success = False
        print(f"{package} installed.")

    return pkg_success


# pylint: enable=not-an-iterable, too-many-branches


# pylint: disable=invalid-name
@export
def md(string: str, styles: Union[str, Iterable[str]] = None):
    """
    Return string as Markdown with optional style.

    Parameters
    ----------
    string : str
        The string to display
    styles : Union[str, Iterable[str]], optional
        A style mnemonic or collection of styles. If multiple styles,
        these can be supplied as an interable of strings or a comma-separated
        string, by default None

    """
    style_str = ""
    if isinstance(styles, str):
        if "," in styles:
            styles = [style.strip() for style in styles.split(",")]
        else:
            style_str = _F_STYLES.get(styles, "")
    if isinstance(styles, list):
        style_str = ";".join([_F_STYLES.get(style, "") for style in styles])
    display(HTML(f"<p style='{style_str}'>{string}</p>"))


# pylint: enable=invalid-name


@export
def md_warn(string: str):
    """
    Return string as a warning - orange text prefixed by "Warning".

    Parameters
    ----------
    string : str
        The warning message.

    """
    md(f"Warning: {string}", "bold, orange, large")


@export
def md_error(string: str):
    """
    Return string as an error - red text prefixed by "Error".

    Parameters
    ----------
    string : str
        The error message.

    """
    md(f"Error: {string}", "bold, orange, large")


# Styles available to use in the above Markdown tools.
_F_STYLES = {
    "bold": "font-weight: bold",
    "italic": "font-style: italic",
    "red": "color: red",
    "green": "color: green",
    "blue": "color: blue",
    "large": "font-size: 130%",
    "heading": "font-size: 200%",
}


@export
def is_ipython() -> bool:
    """
    Return True if running in IPython environment.

    Returns
    -------
    bool
        True if running in IPython environment,
        otherwise False

    """
    return bool(get_ipython())


def check_kwarg(arg_name: str, legal_args: List[str]):
    """
    Check argument names against a list.

    Parameters
    ----------
    arg_name : str
        Argument to check
    legal_args : List[str]
        List of possible arguments.

    Raises
    ------
    NameError
        If the argument is not legal. If the `arg_name` is
        a close match to one or more, `legal_args` these are
        returned in the exception.

    """
    if arg_name not in legal_args:
        closest = difflib.get_close_matches(arg_name, legal_args)
        mssg = f"{arg_name} is not a recognized argument. "
        if len(closest) == 1:
            mssg += f"Closest match is '{closest[0]}'"
        elif closest:
            match_list = [f"'{mtch}'" for mtch in closest]
            mssg += f"Closest matches are {', '.join(match_list)}"
        else:
            mssg += f"Valid arguments are {', '.join(legal_args)}"
        raise NameError(arg_name, mssg)


def check_kwargs(supplied_args: Dict[str, Any], legal_args: List[str]):
    """
    Check all kwargs names against a list.

    Parameters
    ----------
    supplied_args : Dict[str, Any]
        Arguments to check
    legal_args : List[str]
        List of possible arguments.

    Raises
    ------
    NameError
        If any of the arguments are not legal. If the an arg is
        a close match to one or more `legal_args`, these are
        returned in the exception.

    """
    name_errs = []
    for name in supplied_args:
        try:
            check_kwarg(name, legal_args)
        except NameError as err:
            name_errs.append(err)
    if name_errs:
        raise NameError(name_errs)


_U_TEST_ENV = "MP_UNIT_TEST"


def unit_testing() -> bool:
    """
    Return True if in unit testing.

    Returns
    -------
    bool
        True if in unit testing

    """
    return _U_TEST_ENV in os.environ


# pylint: disable=invalid-name
def set_unit_testing(on: bool = True):
    """
    Set flag env var to indicated that code is being unit-tested.

    Parameters
    ----------
    on : bool, optional
        Turn unit testing flag on or off, by default True

    """
    if on:
        os.environ[_U_TEST_ENV] = "True"
    else:
        os.environ.pop(_U_TEST_ENV, None)


# pylint: enable=invalid-name


def is_valid_uuid(uuid_str: Any) -> bool:
    """
    Return true if `uuid_str` is a value GUID/UUID.

    Parameters
    ----------
    uuid_str : Any
        String to test

    Returns
    -------
    bool
        True if valid GUID/UUID.

    """
    if not uuid_str:
        return False
    try:
        uuid.UUID(uuid_str)
    except (ValueError, TypeError):
        return False
    return True


def valid_pyname(identifier: str) -> str:
    """
    Return legal Python identifier, which doesn't collide with builtins.

    Parameters
    ----------
    identifier : str
        The input identifier

    Returns
    -------
    str
        The cleaned identifier

    """
    builtin_names = set(dir(builtins))
    if identifier in builtin_names:
        identifier = f"{identifier}_bi"
    identifier = re.sub("[^a-zA-Z0-9_]", "_", identifier)
    if identifier[0].isdigit():
        identifier = f"n_{identifier}"
    return identifier



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Splunk Uploader class."""
from pathlib import Path
from typing import Any
from tqdm.notebook import tqdm
import pandas as pd
from pandas.io.parsers import ParserError

from .uploader_base import UploaderBase
from ..._version import VERSION
from ..drivers.splunk_driver import SplunkDriver
from ...common.exceptions import MsticpyConnectionError, MsticpyUserError

__version__ = VERSION
__author__ = "Pete Bryan"


class SplunkUploader(UploaderBase):
    """Uploader class for Splunk."""

    def __init__(self, username: str, host: str, password: str, **kwargs):
        """Initialize a Splunk Uploader instance."""
        super().__init__()
        self._kwargs = kwargs
        self.workspace = host
        self.workspace_secret = password
        self.user = username
        self.driver = SplunkDriver()
        self.port = kwargs.get("port", 8089)
        self._debug = kwargs.get("debug", False)
        self._connect = kwargs.get("connect", True)
        self.connected = False
        if self._connect:
            self.connect()

    def connect(self):
        """Connect to Splunk host."""
        self.driver.connect(
            host=self.workspace,
            username=self.user,
            password=self.workspace_secret,
            port=self.port,
        )
        self.connected = True

    def _post_data(
        self,
        data: pd.DataFrame,
        index_name: str,
        table_name: Any,
        host: str = None,
        **kwargs,
    ):
        """
        Write data to the Splunk instance connected to.

        Parameters
        ----------
        data : pd.DataFrame
            Data to upload.
        index_name : str
            Name of the Splunk Index to add data to.
        table_name : str
            The souretype in Splunk data will be uploaded to.
        host : str, optional
            The hostname associated with the uploaded data, by default "Upload".

        """
        if not self.connected:
            raise MsticpyConnectionError(
                "Splunk host not connected, please call .connect before proceding.",
                title="Splunk host not connected",
            )
        if not host:
            host = "Upload"
        create_idx = kwargs.get("create_index", False)
        index = self._load_index(index_name, create_idx)
        progress = tqdm(total=len(data.index), desc="Rows", position=0)
        for row in data.iterrows():
            data = row[1].to_csv()
            try:
                data.encode(encoding="latin-1")
            except UnicodeEncodeError:
                data = data.encode(encoding="utf-8")
            index.submit(data, sourcetype=table_name, host=host)
            progress.update(1)
        progress.close()
        if self._debug is True:
            print("Upload complete")

    # pylint: disable=arguments-differ
    def upload_df(  # type: ignore
        self,
        data: pd.DataFrame,
        table_name: str,
        index_name: str,
        create_index: bool = False,
        **kwargs,
    ):
        """
        Upload a Pandas DataFrame to Splunk.

        Parameters
        ----------
        data : pd.DataFrame
            Data to upload.
        table_name : str
            The souretype in Splunk data will be uploaded to.
        index_name : str
            Name of the Splunk Index to add data to.
        host : str, optional
            Host name to upload data with, default will be 'Upload'
        create_index : bool, optional
            Set this to true to create the index if it doesn't already exist. Default is False.

        """
        host = kwargs.get("host", None)
        if not isinstance(data, pd.DataFrame):
            raise MsticpyUserError(
                "Data must be in Pandas DataFrame format.",
                title="incorrect data format",
            )
        self._post_data(
            data=data,
            table_name=table_name,
            index_name=index_name,
            create_index=create_index,
            host=host,
        )

    def upload_file(  # type: ignore
        self,
        file_path: str,
        index_name: str,
        table_name: str = None,
        delim: str = ",",
        create_index=False,
        **kwargs,
    ):
        """
        Upload a seperated value file to Splunk.

        Parameters
        ----------
        file_path : str
            Path to the file to upload.
        index_name : str
            Name of the Splunk Index to add data to.
        table_name : str, optional
            The souretype in Splunk data will be uploaded to, if not set the file name will be used.
        delim : str, optional
            Seperator value in file, by default ","
        host : str, optional
            Host name to upload data with, default will be 'Upload'
        create_index : bool, optional
            Set this to true to create the index if it doesn't already exist. Default is False.

        """
        host = kwargs.get("host", None)
        path = Path(file_path)
        try:
            data = pd.read_csv(path, delimiter=delim)
        except (ParserError, UnicodeDecodeError) as parse_err:
            raise MsticpyUserError(
                "The file specified is not a seperated value file.",
                "Incorrect file type.",
            ) from parse_err

        if not table_name:
            table_name = path.stem
        self._post_data(
            data=data,
            table_name=table_name,
            index_name=index_name,
            host=host,
            create_index=create_index,
        )

    def upload_folder(  # type: ignore
        self,
        folder_path: str,
        index_name: str,
        table_name: str = None,
        delim: str = ",",
        create_index=False,
        **kwargs,
    ):
        """
        Upload all files in a folder to Splunk.

        Parameters
        ----------
        folder_path : str
            Path to folder to upload.
        index_name : str
            Name of the Splunk Index to add data to, if it doesn't exist it will be created.
        table_name : str, optional
            The souretype in Splunk data will be uploaded to, if not set the file name will be used.
        delim : str, optional
            Seperator value in files, by default ","
        host : str, optional
            Host name to upload data with, default will be 'Upload'
        create_index : bool, optional
            Set this to true to create the index if it doesn't already exist. Default is False.

        """
        host = kwargs.get("host", None)
        glob_pat = kwargs.get("glob", "*")
        t_name = bool(table_name)
        input_files = Path(folder_path).glob(glob_pat)
        f_progress = tqdm(total=len(list(input_files)), desc="Files", position=0)
        for path in input_files:
            try:
                data = pd.read_csv(path, delimiter=delim)
            except (ParserError, UnicodeDecodeError) as parse_err:
                raise MsticpyUserError(
                    "The file specified is not a seperated value file.",
                    title="Incorrect file type.",
                ) from parse_err
            if not t_name:
                table_name = path.stem
            self._post_data(
                data=data,
                table_name=table_name,
                index_name=index_name,
                host=host,
                create_index=create_index,
            )
            f_progress.update(1)
            if self._debug is True:
                print(f"{str(path)} uploaded to {table_name}")
        f_progress.close()

    # pylint: enable=arguments-differ

    def _check_index(self, index_name: str):
        """Check if index exists in Splunk host."""
        service_list = [item.name for item in self.driver.service.indexes]
        if index_name in service_list:
            return True
        return False

    def _load_index(self, index_name, create: bool = True):
        """Load specified Index or create if it doesn't exist."""
        if self._check_index(index_name):
            return self.driver.service.indexes[index_name]
        if not self._check_index(index_name) and create:
            return self.driver.service.indexes.create(index_name)

        raise MsticpyConnectionError("Index not present in Splunk host.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
IBM XForce Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from typing import Any, Tuple

import attr

from .ti_provider_base import LookupResult, TISeverity
from .http_base import HttpProvider, IoCLookupParams
from ...common.utility import export
from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
@attr.s
class _XForceParams(IoCLookupParams):
    # override IoCLookupParams to set common defaults
    def __attrs_post_init__(self):
        self.auth_str = ["{API_ID}", "{API_KEY}"]
        self.auth_type = "HTTPBasic"


@export
class XForce(HttpProvider):
    """IBM XForce Lookup."""

    _BASE_URL = "https://api.xforce.ibmcloud.com"

    _IOC_QUERIES = {
        "ipv4": _XForceParams(path="/ipr/{observable}"),
        "ipv4-rep": _XForceParams(path="/ipr/history/{observable}"),
        "ipv4-malware": _XForceParams(path="/ipr/malware/{observable}"),
        "ipv4-whois": _XForceParams(path="/whois/{observable}"),
        "dns-passivedns": _XForceParams(path="/resolve/{observable}"),
        "hostname-whois": _XForceParams(path="/whois/{observable}"),
        "file_hash": _XForceParams(path="/malware/{observable}"),
        "url": _XForceParams(path="/url/{observable}"),
        "url-malware": _XForceParams(path="/url/malware/{observable}"),
    }

    # pylint: disable=duplicate-code
    # aliases
    _IOC_QUERIES["ipv6"] = _IOC_QUERIES["ipv4"]
    _IOC_QUERIES["ipv6-rep"] = _IOC_QUERIES["ipv4-rep"]
    _IOC_QUERIES["ipv6-malware"] = _IOC_QUERIES["ipv4-malware"]
    _IOC_QUERIES["ipv6-whois"] = _IOC_QUERIES["ipv4-whois"]
    _IOC_QUERIES["md5_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha1_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["sha256_hash"] = _IOC_QUERIES["file_hash"]
    _IOC_QUERIES["dns"] = _IOC_QUERIES["url"]
    _IOC_QUERIES["dns-malware"] = _IOC_QUERIES["url-malware"]
    _IOC_QUERIES["ipv4-passivedns"] = _IOC_QUERIES["dns-passivedns"]
    _IOC_QUERIES["ipv6-passivedns"] = _IOC_QUERIES["dns-passivedns"]
    _IOC_QUERIES["hostname-whois"] = _IOC_QUERIES["ipv4-whois"]
    _IOC_QUERIES["dns-whois"] = _IOC_QUERIES["ipv4-whois"]

    _REQUIRED_PARAMS = ["API_ID", "API_KEY"]

    # pylint: disable=too-many-branches
    def parse_results(self, response: LookupResult) -> Tuple[bool, TISeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : LookupResult
            The returned data response

        Returns
        -------
        Tuple[bool, TISeverity, Any]
            bool = positive or negative hit
            TISeverity = enumeration of severity
            Object with match details

        """
        severity = TISeverity.information
        if self._failed_response(response) or not isinstance(response.raw_result, dict):
            return False, severity, "Not found."
        result = True
        result_dict = {}
        if (
            response.ioc_type in ["ipv4", "ipv6", "url", "dns"]
            and not response.query_subtype
        ):
            score = response.raw_result.get("score", 0)
            result_dict.update(
                {
                    "score": response.raw_result.get("score", 0),
                    "cats": response.raw_result.get("cats"),
                    "categoryDescriptions": response.raw_result.get(
                        "categoryDescriptions"
                    ),
                    "reason": response.raw_result.get("reason"),
                    "reasonDescription": response.raw_result.get(
                        "reasonDescription", 0
                    ),
                    "tags": response.raw_result.get("tags", 0),
                }
            )
            severity = (
                TISeverity.information
                if score == 0
                else TISeverity.warning
                if score == 1
                else TISeverity.high
            )
        if (
            response.ioc_type in ["file_hash", "md5_hash", "sha1_hash", "sha256_hash"]
            or response.query_subtype == "malware"
        ):
            malware = response.raw_result.get("malware")
            if malware:
                result_dict.update(
                    {
                        "risk": malware.get("risk"),
                        "family": malware.get("family"),
                        "reasonDescription": response.raw_result.get(
                            "reasonDescription", 0
                        ),
                    }
                )
                severity = TISeverity.high
        if (
            response.ioc_type
            in [
                "dns",
                "ipv4",
                "ipv6",
                "hostname",
            ]
            and response.query_subtype in ["info", "passivedns", "whois"]
        ):
            records = response.raw_result.get("total_rows", 0)
            contact = response.raw_result.get("contact", 0)
            if records:
                result_dict.update({"records": records})
            elif contact:
                result_dict.update({"contact": contact})
        return result, severity, result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
entityschema module.

Module for V3 Entities class
"""
import pprint
from abc import ABC, abstractmethod
from enum import Enum
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, Dict, Mapping, Type, Union

from .._version import VERSION
from ..common.utility import export

__version__ = VERSION
__author__ = "Ian Hellen"


_ENTITY_ENUMS: Dict[str, Type] = {}


# pylint: disable=too-many-lines, invalid-name
# pylint: disable=too-many-instance-attributes
@export
class Entity(ABC):
    """
    Entity abstract base class.

    Implements common methods for Entity classes
    """

    ENTITY_NAME_MAP: Dict[str, type] = {}
    _entity_schema: Dict[str, Any] = {}

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of an entity.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            If src_entity is supplied it attempts to extract common
            properties from the source entity and assign them to
            the new instance. (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Type = type(self).__name__.lower()
        # If we have an unknown entity see if we a type passed in
        if self.Type == "unknownentity" and "Type" in kwargs:
            self.Type = kwargs["Type"]
        # Make sure Type is in the class schema dictionary
        self._entity_schema["Type"] = None

        # if we didn't populate AdditionalData, add an empty dict in case it's
        # needed
        if "AdditionalData" not in self:
            self["AdditionalData"] = {}

        if src_entity is not None:
            self._extract_src_entity(src_entity)
            # add AdditionalData dictionary if it's populated
            if "AdditionalData" in src_entity:
                self["AdditionalData"] = src_entity["AdditionalData"]

        if kwargs:
            self.__dict__.update(kwargs)

    def _extract_src_entity(self, src_entity: Mapping[str, Any]):
        """
        Extract source entity properties.

        Parameters
        ----------
        src_entity : Mapping[str, Any]
            The source mappable object from which to
            extract entity properties.

        """
        schema_dict = dict(**(self._entity_schema))
        schema_dict["Type"] = None
        for k, v in schema_dict.items():
            if k not in src_entity:
                continue
            self[k] = src_entity[k]

            if v is not None:
                try:
                    # If the property is an enum
                    if v in _ENTITY_ENUMS:
                        self[k] = _ENTITY_ENUMS[v][src_entity[k]]
                        continue
                except KeyError:
                    # Catch key errors from invalid enum values
                    self[k] = None

                if isinstance(v, tuple):
                    # if the property is a collection
                    entity_list = []
                    for col_entity in src_entity[k]:
                        entity_list.append(Entity.instantiate_entity(col_entity))
                    self[k] = entity_list
                else:
                    # else try to instantiate an entity
                    self[k] = Entity.instantiate_entity(src_entity[k])

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._entity_schema:
            return None
        raise KeyError

    def __setitem__(self, key: str, value: Any):
        """Allow property set using dictionary key syntax."""
        self.__dict__[key] = value

    def __contains__(self, key: str):
        """Allow property in test."""
        # In operator overload
        return key in self.__dict__

    def __getattr__(self, name: str):
        """Return the value of the named property 'name'."""
        if name in self._entity_schema:
            return None
        raise AttributeError(f"{name} is not a valid attribute.")

    def __iter__(self):
        """Iterate over entity_properties."""
        return iter(self.properties)

    def __len__(self) -> int:
        """Return length/number of entity_properties."""
        return len(self.properties)

    def __str__(self) -> str:
        """Return string representation of entity."""
        return pprint.pformat(self._to_dict(self), indent=2, width=100)

    def __repr__(self) -> str:
        """Return repr of entity."""
        params = ", ".join(
            [f"{name}={val}" for name, val in self.properties.items() if val]
        )
        if len(params) > 80:
            params = params[:80] + "..."
        return f"{self.__class__.__name__}({params})"

    def _to_dict(self, entity) -> dict:
        """Return as simple nested dictionary."""
        ent_dict = {}
        for prop, val in entity.properties.items():
            if val is not None:
                ent_dict[prop] = self._to_dict(val) if isinstance(val, Entity) else val
        return ent_dict

    def _repr_html_(self) -> str:
        """
        Display entity in IPython/Notebook.

        Returns
        -------
        HTML
            IPython HTML object

        """
        return self.to_html()

    def to_html(self) -> str:
        """
        Return HTML representation of entity.

        Returns
        -------
        str
            HTML representation of entity

        """
        e_text = str(self)
        e_type = self.Type
        e_text = e_text.replace("\n", "<br>").replace(" ", "&nbsp;")
        return f"<h3>{e_type}</h3>{e_text}"

    @property
    def properties(self) -> dict:
        """
        Return dictionary properties of entity.

        Returns
        -------
        dict
            Entity properties.

        """
        return {
            name: value
            for name, value in self.__dict__.items()
            if not name.startswith("_")
        }

    @property
    @abstractmethod
    def description_str(self) -> str:
        """
        Return Entity Description.

        Returns
        -------
        str
            Entity description (optional). If not overridden
            by the Entity instance type, it will return the
            Type string.

        """
        return self.Type

    # pylint: disable=too-many-branches
    @classmethod
    def instantiate_entity(  # noqa: C901
        cls, raw_entity: Mapping[str, Any]
    ) -> Union["Entity", Mapping[str, Any]]:
        """
        Class factory to return entity from raw dictionary representation.

        Parameters
        ----------
        raw_entity : Mapping[str, Any]
            A mapping object (e.g. dictionary or pandas Series)
            that contains the properties of the entity.

        Returns
        -------
        Entity
            The instantiated entity

        """
        if "Type" not in raw_entity:
            return raw_entity

        entity_type = raw_entity["Type"]

        # We get an undefined-variable warning here. _ENTITY_NAME_MAP
        # is not defined/populated until end of module since it needs
        # entity
        if entity_type in cls.ENTITY_NAME_MAP:
            return cls.ENTITY_NAME_MAP[entity_type](raw_entity)

        raise TypeError("Could not find a suitable type for {}".format(entity_type))


@export
class Account(Entity):
    """
    Account Entity class.

    Attributes
    ----------
    Name : str
        Account Name
    NTDomain : str
        Account NTDomain
    UPNSuffix : str
        Account UPNSuffix
    Host : Host
        Account Host
    LogonId : str
        Account LogonId (deprecated)
    Sid : str
        Account Sid
    AadTenantId : str
        Account AadTenantId
    AadUserId : str
        Account AadUserId
    PUID : str
        Account PUID
    IsDomainJoined : bool
        Account IsDomainJoined
    DisplayName : str
        Account DisplayName

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role: str = "subject",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing Account entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'subject' or 'target' - only relevant if the entity
            is being constructed from an event.
            (the default is 'subject')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        # pylint: disable=locally-disabled, line-too-long
        super().__init__(src_entity=src_entity, **kwargs)
        if src_event is not None:
            if role == "subject" and "SubjectUserName" in src_event:
                self.Name = src_event["SubjectUserName"]
                self.NTDomain = (
                    src_event["SubjectUserDomain"]
                    if "SubjectUserDomain" in src_event
                    else None
                )
                self.Sid = (
                    src_event["SubjectUserSid"]
                    if "SubjectUserSid" in src_event
                    else None
                )
                self.LogonId = (
                    src_event["SubjectLogonId"]
                    if "SubjectLogonId" in src_event
                    else None
                )
            if role == "target" and "TargetUserName" in src_event:
                self.Name = src_event["TargetUserName"]
                self.NTDomain = (
                    src_event["TargetUserDomain"]
                    if "TargetUserDomain" in src_event
                    else None
                )
                self.Sid = (
                    src_event["TargetUserSid"] if "TargetUserSid" in src_event else None
                )
                self.LogonId = (
                    src_event["TargetLogonId"] if "TargetLogonId" in src_event else None
                )

            self.AadTenantId = (
                src_event["AadTenantId"] if "AadTenantId" in src_event else None
            )
            self.AadUserId = (
                src_event["AadUserId"] if "AadUserId" in src_event else None
            )
            self.PUID = src_event["PUID"] if "PUID" in src_event else None
            self.DisplayName = (
                src_event["DisplayName"] if "DisplayName" in src_event else None
            )
            self.UPNSuffix = (
                src_event["UPNSuffix"] if "UPNSuffix" in src_event else None
            )

    # pylint: enable=locally-disabled, line-too-long

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.qualified_name

    @property
    def qualified_name(self) -> str:
        """Windows qualified account name."""
        if "Name" not in self:
            return ""
        name = self["Name"]
        if "NTDomain" in self and self.NTDomain:
            return "{}\\{}".format(self.NTDomain, name)
        if "UPNSuffix" in self and self.UPNSuffix:
            return "{}@{}".format(name, self.UPNSuffix)
        if "Host" in self and self.Host:
            return "{}\\{}".format(self.Host.HostName, name)
        return name

    _entity_schema = {
        # Name (type System.String)
        "Name": None,
        # NTDomain (type System.String)
        "NTDomain": None,
        # UPNSuffix (type System.String)
        "UPNSuffix": None,
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": "Host",
        # LogonId (type System.String)
        "LogonId": None,
        # Sid (type System.String)
        "Sid": None,
        # AadTenantId (type System.Nullable`1[System.Guid])
        "AadTenantId": None,
        # AadUserId (type System.Nullable`1[System.Guid])
        "AadUserId": None,
        # PUID (type System.Nullable`1[System.Guid])
        "PUID": None,
        # IsDomainJoined (type System.Nullable`1[System.Boolean])
        "IsDomainJoined": None,
        # DisplayName (type System.String)
        "DisplayName": None,
    }


@export
class SecurityGroup(Entity):
    """
    SecurityGroup Entity class.

    Attributes
    ----------
    DistinguishedName : str
        SecurityGroup DistinguishedName
    SID : str
        SecurityGroup SID
    ObjectGuid : str
        SecurityGroup ObjectGuid

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self):
        """Return Entity Description."""
        return self.DistinguishedName

    _entity_schema = {
        # DistinguishedName (type System.String)
        "DistinguishedName": None,
        # SID (type System.String)
        "SID": None,
        # ObjectGuid (type System.String)
        "ObjectGuid": None,
    }


@export
class HostLogonSession(Entity):
    """
    HostLogonSession Entity class.

    Attributes
    ----------
    Account : Account
        HostLogonSession Account
    StartTimeUtc : datetime
        HostLogonSession StartTimeUtc
    EndTimeUtc : datetime
        HostLogonSession EndTimeUtc
    Host : Host
        HostLogonSession Host
    SessionId : str
        HostLogonSession SessionId

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.


        """
        super().__init__(src_entity=src_entity, **kwargs)

        if src_event is not None:
            if "TimeCreatedUtc" in src_event:
                self.StartTimeUtc = src_event["TimeCreatedUtc"]
            elif "TimeGenerated" in src_event:
                self.StartTimeUtc = src_event["TimeGenerated"]
            self.EndTimeUtc = self.StartTimeUtc
            self.SessionId = (
                src_event["TargetLogonId"] if "TargetLogonId" in src_event else None
            )

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Host.HostName}: session: {self.SessionId}"

    _entity_schema = {
        # Account
        "Account": "Account",
        # StartTimeUtc (type System.Nullable`1[System.DateTime])
        "StartTimeUtc": None,
        # EndTimeUtc (type System.Nullable`1[System.DateTime])
        "EndTimeUtc": None,
        # Host
        "Host": "Host",
        # SessionId (type System.String)
        "SessionId": None,
    }


@export
class CloudApplication(Entity):
    """
    CloudApplication Entity class.

    Attributes
    ----------
    Name : str
        CloudApplication Name

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.Name

    _entity_schema = {
        # Name (type System.String)
        "Name": None
    }


@export
class DnsResolve(Entity):
    """
    DNS Resolve Entity class.

    Attributes
    ----------
    DomainName : str
        DnsResolve DomainName
    IpAdresses : List[str]
        DnsResolve IpAdresses
    DnsServerIp : IPAddress
        DnsResolve DnsServerIp
    HostIpAddress : IPAddress
        DnsResolve HostIpAddress

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.DomainName}: IPs: {repr(self.IpAdresses)}"

    _entity_schema = {
        # DomainName (type System.String)
        "DomainName": None,
        # IpAdresses (type System.Collections.Generic.List`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.IP])
        "IpAdresses": None,
        # DnsServerIp (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "DnsServerIp": "IPAddress",
        # HostIpAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "HostIpAddress": "IPAddress",
    }


@export
class File(Entity):
    """
    File Entity class.

    Attributes
    ----------
    FullPath : str
        File FullPath
    Directory : str
        File Directory
    Name : str
        File Name
    Md5 : str
        File Md5
    Host : str
        File Host
    Sha1 : str
        File Sha1
    Sha256 : str
        File Sha256
    Sha256Ac : str
        File Sha256Ac
    FileHashes : List[FileHash]
        File FileHashes

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role: str = "new",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'new' or 'parent' - only relevant if the entity
            is being constructed from an event.
            (the default is 'new')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

        if src_event is not None:
            if role == "new" and "NewProcessName" in src_event:
                self._add_paths(src_event["NewProcessName"])
            elif role == "parent" and "ParentProcessName" in src_event:
                self._add_paths(src_event["ParentProcessName"])

        if "FullPath" not in self:
            file = self["Name"]
            directory = self["Directory"]
            sep = self.path_separator if directory else None
            self["FullPath"] = f"{directory}{sep}{file}"

    @property
    def path_separator(self):
        """Return the path separator used by the file."""
        directory = self["Directory"]
        if directory and "/" in directory:
            return "/"
        return "\\"

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.FullPath

    _entity_schema = {
        # FullPath (type System.String)
        "FullPath": None,
        # Directory (type System.String)
        "Directory": None,
        # Name (type System.String)
        "Name": None,
        # Md5 (type System.String)
        "Md5": None,
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": None,
        # Sha1 (type System.String)
        "Sha1": None,
        # Sha256 (type System.String)
        "Sha256": None,
        # Sha256Ac (type System.String)
        "Sha256Ac": None,
        "FileHashes": (list, "FileHash"),
    }

    def _add_paths(self, full_path):
        if "/" in full_path:
            self.PathSeparator = "/"
            self.OSFamily = OSFamily.Linux
        else:
            self.PathSeparator = "\\"
            self.OSFamily = OSFamily.Windows

        self.FullPath = full_path
        self.Name = full_path.split(self.PathSeparator)[-1]
        self.Directory = full_path.split(self.PathSeparator)[:-1]


@export
class FileHash(Entity):
    """
    File Hash class.

    Attributes
    ----------
    Algorithm : Algorithm
        FileHash Algorithm
    Value : str
        FileHash Value


    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Algorithm}: {self.Value}"

    _entity_schema = {
        # The hash algorithm (type System.String)
        "Algorithm": "Algorithm",
        # Value (type System.String)
        "Value": None,
    }


@export
class Algorithm(Enum):
    """FileHash Algorithm Enumeration."""

    Unknown = 0
    MD5 = 1
    SHA1 = 2
    SHA256 = 3
    SHA256AC = 4


_ENTITY_ENUMS[Algorithm.__name__] = Algorithm


@export
class Host(Entity):
    """
    Host Entity class.

    Attributes
    ----------
    DnsDomain : str
        Host DnsDomain
    NTDomain : str
        Host NTDomain
    HostName : str
        Host HostName
    NetBiosName : str
        Host NetBiosName
    AzureID : str
        Host AzureID
    OMSAgentID : str
        Host OMSAgentID
    OSFamily : str
        Host OSFamily
    IsDomainJoined : bool
        Host IsDomainJoined

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)
        self._computer = None
        if src_event is not None and "Computer" in src_event:
            self._computer = src_event["Computer"]
            if "." in src_event["Computer"]:
                self.HostName = src_event["Computer"].split(".", 1)[0]
                self.DnsDomain = src_event["Computer"].split(".", 1)[1]
            else:
                self.HostName = src_event["Computer"]
            self.NetBiosName = self.HostName

    @property
    def computer(self) -> str:
        """Return computer from source event."""
        return self._computer if self._computer is not None else self.fqdn

    @property
    def fqdn(self) -> str:
        """Construct FQDN from host + dns."""
        if self.DnsDomain:
            return f"{self.HostName}.{self.DnsDomain}"
        return self.HostName

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.fqdn} ({self.OSFamily})"

    _entity_schema = {
        # DnsDomain (type System.String)
        "DnsDomain": None,
        # NTDomain (type System.String)
        "NTDomain": None,
        # HostName (type System.String)
        "HostName": None,
        # NetBiosName (type System.String)
        "NetBiosName": None,
        # AzureID (type System.String)
        "AzureID": None,
        # OMSAgentID (type System.String)
        "OMSAgentID": None,
        # OSFamily (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.OSFamily])
        "OSFamily": None,
        # IsDomainJoined (type System.Nullable`1[System.Boolean])
        "IsDomainJoined": None,
    }


@export
class IpAddress(Entity):
    """
    IPAddress Entity class.

    Attributes
    ----------
    Address : str
        IpAddress Address
    Location : GeoLocation
        IpAddress Location
    ThreatIntelligence : List[ThreatIntelligence]
        IpAddress ThreatIntelligence

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

        if src_event is not None and "IpAddress" in src_event:
            self.Address = src_event["IpAddress"]

    @property
    def ip_address(self) -> Union[IPv4Address, IPv6Address]:
        """Return a python ipaddress object from the entity property."""
        return ip_address(self["Address"])

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.Address

    _entity_schema = {
        # Address (type System.String)
        "Address": None,
        # Location (type Microsoft.Azure.Security.Detection.AlertContracts
        # .V3.ContextObjects.GeoLocation)
        "Location": "GeoLocation",
        # ThreatIntelligence (type System.Collections.Generic.List`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3
        # .ContextObjects.ThreatIntelligence])
        "ThreatIntelligence": (list, "Threatintelligence"),
    }


@export
class GeoLocation(Entity):
    """
    GeoLocation class.

    Attributes
    ----------
    CountryCode : str
        GeoLocation CountryCode
    CountryName : str
        GeoLocation CountryName
    State : str
        GeoLocation State
    City : str
        GeoLocation City
    Longitude : float
        GeoLocation Longitude
    Latitude : float
        GeoLocation Latitude
    Asn : str
        GeoLocation Asn

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.CountryCode}; {self.State}; {self.City}"

    _entity_schema = {
        # str
        "CountryCode": None,
        # str
        "CountryName": None,
        # str
        "State": None,
        # str
        "City": None,
        # double?
        "Longitude": None,
        # double?
        "Latitude": None,
        # int
        "Asn": None,
    }


@export
class Malware(Entity):
    """
    Malware Entity class.

    Attributes
    ----------
    Name : str
        Malware Name
    Category : str
        Malware Category
    File : File
        Malware File
    Files : List[File]
        Malware Files
    Processes : List[Process]
        Malware Processes

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Name}: {self.Category}"

    _entity_schema = {
        # Name (type System.String)
        "Name": None,
        # Category (type System.String)
        "Category": None,
        # File (type Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.File)
        "File": "File",
        "Files": (list, "File"),
        "Processes": (list, "Process"),
    }


@export
class NetworkConnection(Entity):
    """
    NetworkConnection Entity class.

    Attributes
    ----------
    SourceAddress : IPAddress
        NetworkConnection SourceAddress
    SourcePort : int
        NetworkConnection SourcePort
    DestinationAddress : IPAddress
        NetworkConnection DestinationAddress
    DestinationPort : int
        NetworkConnection DestinationPort
    Protocol : str
        NetworkConnection Protocol


    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return "{}:{} [{}]-> {}:{}".format(
            self.SourceAddress,
            self.SourcePort,
            self.Protocol,
            self.DestinationAddress,
            self.DestinationPort,
        )

    _entity_schema = {
        # SourceAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "SourceAddress": "IPAddress",
        # SourcePort (type System.Nullable`1[System.Int32])
        "SourcePort": None,
        # DestinationAddress (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.IP)
        "DestinationAddress": "IPAddress",
        # DestinationPort (type System.Nullable`1[System.Int32])
        "DestinationPort": None,
        # Protocol (type System.Nullable`1[System.Net.Sockets.ProtocolType])
        "Protocol": None,
    }


@export
class Process(Entity):
    """
    Process Entity class.

    Attributes
    ----------
    ProcessId : str
        Process ProcessId
    CommandLine : str
        Process CommandLine
    ElevationToken : str
        Process ElevationToken
    CreationTimeUtc : datetime
        Process CreationTimeUtc
    ImageFile : File
        Process ImageFile
    Account : Account
        Process Account
    ParentProcess : Process
        Process ParentProcess
    Host : Host
        Process Host
    LogonSession : HostLogonSession
        Process LogonSession

    """

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role="new",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'new' or 'parent' - only relevant if the entity
            is being constructed from an event.
            (the default is 'new')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)
        # pylint: disable=locally-disabled, line-too-long
        if src_event is not None:
            if role == "new":
                self.ProcessId = (
                    src_event["NewProcessId"] if "NewProcessId" in src_event else None
                )
                self.CommandLine = (
                    src_event["CommandLine"] if "CommandLine" in src_event else None
                )
                if "TimeCreatedUtc" in src_event:
                    self.CreationTimeUtc = src_event["TimeCreatedUtc"]
                elif "TimeGenerated" in src_event:
                    self.CreationTimeUtc = src_event["TimeGenerated"]
                self.ProcessId = (
                    src_event["NewProcessId"] if "NewProcessId" in src_event else None
                )
                self.ImageFile = File(src_event=src_event, role="new")
                self.Account = Account(src_event=src_event, role="subject")

                if "ParentProcessName" in src_event or "ProcessName" in src_event:
                    parent = Process(src_event=src_event, role="parent")
                    self.ParentProcess = parent

                # Linux properties
                self.success = src_event["success"] if "success" in src_event else None
                self.audit_user = (
                    src_event["audit_user"] if "audit_user" in src_event else None
                )
                self.auid = src_event["auid"] if "auid" in src_event else None
                self.group = src_event["group"] if "group" in src_event else None
                self.gid = src_event["gid"] if "gid" in src_event else None
                self.effective_user = (
                    src_event["effective_user"]
                    if "effective_user" in src_event
                    else None
                )
                self.euid = src_event["euid"] if "euid" in src_event else None
                self.effective_group = (
                    src_event["effective_group"]
                    if "effective_group" in src_event
                    else None
                )
                self.egid = (
                    src_event["effective_group"]
                    if "effective_group" in src_event
                    else None
                )
                self.cwd = src_event["cwd"] if "cwd" in src_event else None
                self.name = src_event["cwd"] if "cwd" in src_event else None
            else:
                self.ProcessId = (
                    src_event["ProcessId"] if "ProcessId" in src_event else None
                )
                self.ImageFile = File(src_event=src_event, role="parent")

    # pylint: enable=locally-disabled, line-too-long

    @property
    def ProcessName(self) -> str:  # noqa: N802
        """Return the name of the process file."""
        file = self["ImageFile"]
        return file.Name if file else None

    @property
    def ProcessFilePath(self) -> str:  # noqa: N802
        """Return the name of the process file path."""
        file = self["ImageFile"]
        return file.FullPath if file else None

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.ProcessFilePath}: {self.CommandLine}"

    _entity_schema = {
        # ProcessId (type System.String)
        "ProcessId": None,
        # CommandLine (type System.String)
        "CommandLine": None,
        # ElevationToken (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.ElevationToken])
        "ElevationToken": None,
        # CreationTimeUtc (type System.Nullable`1[System.DateTime])
        "CreationTimeUtc": None,
        # ImageFile (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.File)
        "ImageFile": "File",
        # Account (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Account)
        "Account": "Account",
        # ParentProcess (type Microsoft.Azure.Security.Detection.AlertContracts
        # .V3.Entities.Process)
        "ParentProcess": "Process",
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": "Host",
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.HostLogonSession)
        "LogonSession": "HostLogonSession",
    }


@export
class RegistryHive(Enum):
    """RegistryHive enumeration."""

    # <summary>HKEY_LOCAL_MACHINE</summary>
    HKEY_LOCAL_MACHINE = 0
    # <summary>HKEY_CLASSES_ROOT</summary>
    HKEY_CLASSES_ROOT = 1
    # <summary>HKEY_CURRENT_CONFIG</summary>
    HKEY_CURRENT_CONFIG = 2
    # <summary>HKEY_USERS</summary>
    HKEY_USERS = 3
    # <summary>HKEY_CURRENT_USER_LOCAL_SETTINGS</summary>
    HKEY_CURRENT_USER_LOCAL_SETTINGS = 4
    # <summary>HKEY_PERFORMANCE_DATA</summary>
    HKEY_PERFORMANCE_DATA = 5
    # <summary>HKEY_PERFORMANCE_NLSTEXT</summary>
    HKEY_PERFORMANCE_NLSTEXT = 6
    # <summary>HKEY_PERFORMANCE_TEXT</summary>
    HKEY_PERFORMANCE_TEXT = 7
    # <summary>HKEY_A</summary>
    HKEY_A = 8
    # <summary>HKEY_CURRENT_USER</summary>
    HKEY_CURRENT_USER = 9


_ENTITY_ENUMS[RegistryHive.__name__] = RegistryHive


@export
class RegistryKey(Entity):
    """
    RegistryKey Entity class.

    Attributes
    ----------
    Hive : RegistryHive
        RegistryKey Hive
    Key : str
        RegistryKey Key

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Hive}\\{self.Key}"

    _entity_schema = {
        # Hive (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Entities.RegistryHive])
        "Hive": "RegistryHive",
        # Key (type System.String)
        "Key": None,
    }


class RegistryValue(Entity):
    """
    RegistryValue Entity class.

    Attributes
    ----------
    Key : str
        RegistryValue Key
    Name : str
        RegistryValue Name
    Value : str
        RegistryValue Value
    ValueType : str
        RegistryValue ValueType

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.Name}[{self.ValueType}]:{repr(self.Value)}"

    _entity_schema = {
        # Key (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.RegistryKey)
        "Key": None,
        # Name (type System.String)
        "Name": None,
        # Value (type System.String)
        "Value": None,
        # ValueType (type System.Nullable`1[Microsoft.Win32.RegistryValueKind])
        "ValueType": None,
    }


@export
class OSFamily(Enum):
    """OSFamily enumeration."""

    Linux = 0
    Windows = 1


_ENTITY_ENUMS[OSFamily.__name__] = OSFamily


@export
class ElevationToken(Enum):
    """ElevationToken enumeration."""

    Default = 0
    Full = 1
    Limited = 2


_ENTITY_ENUMS[ElevationToken.__name__] = ElevationToken


@export
class AzureResource(Entity):
    """
    AzureResource Entity class.

    Attributes
    ----------
    ResourceId : str
        AzureResource ResourceId
    SubscriptionId : str
        AzureResource SubscriptionId
    ResourceIdParts : Dict[str, str]
        AzureResource ResourceIdParts


    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.ResourceId

    _entity_schema = {
        # ResourceId (type System.String)
        "ResourceId": None,
        # SubscriptionId (type System.String)
        "SubscriptionId": None,
        # ResourceIdParts (type System.Collections.Generic.IReadOnlyDictionary`2
        # [System.String,System.String])
        "ResourceIdParts": None,
    }


@export
class Alert(Entity):
    """
    Alert Entity class.

    Attributes
    ----------
    DisplayName : str
        Alert DisplayName
    CompromisedEntity : str
        Alert CompromisedEntity
    Count : int
        Alert Count
    StartTimeUtc : datetime
        Alert StartTimeUtc
    EndTimeUtc : datetime
        Alert EndTimeUtc
    Severity : str
        Alert Severity
    SystemAlertIds : List[str]
        Alert SystemAlertIds
    AlertType : str
        Alert AlertType
    VendorName : str
        Alert VendorName
    ProviderName : str
        Alert ProviderName

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.DisplayName} ({self.StartTimeUtc}) {self.CompromisedEntity}"

    _entity_schema = {
        # DisplayName (type System.String)
        "DisplayName": None,
        # CompromisedEntity (type System.String)
        "CompromisedEntity": None,
        # Count (type System.Nullable`1[System.Int32])
        "Count": None,
        # StartTimeUtc (type System.Nullable`1[System.DateTime])
        "StartTimeUtc": None,
        # EndTimeUtc (type System.Nullable`1[System.DateTime])
        "EndTimeUtc": None,
        # Severity (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Severity])
        "Severity": None,
        # SystemAlertIds (type System.Collections.Generic.List`1[System.String])
        "SystemAlertIds": None,
        # AlertType (type System.String)
        "AlertType": None,
        # VendorName (type System.String)
        "VendorName": None,
        # ProviderName (type System.String)
        "ProviderName": None,
    }


@export
class Threatintelligence(Entity):
    """
    Threatintelligence Entity class.

    Attributes
    ----------
    ProviderName : str
        Threatintelligence ProviderName
    ThreatType : str
        Threatintelligence ThreatType
    ThreatName : str
        Threatintelligence ThreatName
    Confidence : str
        Threatintelligence Confidence
    ReportLink : str
        Threatintelligence ReportLink
    ThreatDescription : str
        Threatintelligence ThreatDescription

    """

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

            :param src_entity: instantiate entity using properties of src entity
            :param kwargs: key-value pair representation of entity
        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.DisplayName} ({self.StartTimeUtc}) {self.CompromisedEntity}"

    _entity_schema = {
        # String Name of the provider from whom this
        # Threat Intelligence information was received
        "ProviderName": None,
        "ThreatType": None,
        "ThreatName": None,
        "Confidence": None,
        "ReportLink": None,
        "ThreatDescription": None,
    }


@export
class UnknownEntity(Entity):
    """Generic Entity class."""

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

            :param src_entity: instantiate entity using properties of src entity
            :param kwargs: key-value pair representation of entity
        """
        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return "OtherEntity"

    _entity_schema = {}  # type: Dict[str, Any]


# Dictionary to map text names of types to the class.
Entity.ENTITY_NAME_MAP.update(
    {
        "account": Account,
        "host": Host,
        "process": Process,
        "file": File,
        "cloudapplication": CloudApplication,
        "dnsresolve": DnsResolve,
        "ipaddress": IpAddress,
        "ip": IpAddress,
        "networkconnection": NetworkConnection,
        "malware": Malware,
        "registry-key": RegistryKey,
        "registrykey": RegistryKey,
        "registry-value": RegistryValue,
        "registryvalue": RegistryValue,
        "host-logon-session": HostLogonSession,
        "hostlogonsession": HostLogonSession,
        "filehash": FileHash,
        "security-group": SecurityGroup,
        "securitygroup": SecurityGroup,
        "alerts": Alert,
        "alert": Alert,
    }
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query helper definitions."""
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union

from ..common.utility import export
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class DataFamily(Enum):
    """
    Enumeration of data families.

    Used to identify which queries are relevant for which
    data sources.
    """

    Unknown = 0
    WindowsSecurity = 1
    LinuxSecurity = 2
    SecurityAlert = 3
    SecurityGraphAlert = 4
    LinuxSyslog = 5
    AzureNetwork = 6
    MDATP = 7
    Splunk = 8

    @classmethod
    def parse(cls, value: Union[str, int]) -> "DataFamily":
        """
        Convert string or int to enum.

        Parameters
        ----------
        value : Union[str, int]
            value to parse

        """
        if isinstance(value, cls):
            return value

        parsed_enum = cls.Unknown
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                # match to value if case is incorrect
                for e_name, e_val in cls.__members__.items():
                    if e_name.upper() == value.upper():
                        return e_val
                return cls.Unknown
        if isinstance(value, int):
            try:
                parsed_enum = cls(value)
            except ValueError:
                parsed_enum = cls.Unknown
        return parsed_enum


@export
class DataEnvironment(Enum):
    """
    Enumeration of data environments.

    Used to identify which queries are relevant for which
    data sources.
    """

    Unknown = 0
    LogAnalytics = 1
    AzureSentinel = 1  # alias of LogAnalytics
    Kusto = 2
    AzureSecurityCenter = 3
    SecurityGraph = 4
    MDATP = 5
    LocalData = 6
    Splunk = 7
    Mordor = 8

    @classmethod
    def parse(cls, value: Union[str, int]) -> "DataEnvironment":
        """
        Convert string or int to enum.

        Parameters
        ----------
        value : Union[str, int]
            value to parse

        """
        if isinstance(value, cls):
            return value

        parsed_enum = cls.Unknown
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                pass
        if isinstance(value, int):
            parsed_enum = cls(value)
        return parsed_enum


# pylint: disable=too-few-public-methods
@export
class QueryParamProvider(ABC):
    """
    Abstract type for QueryParamProvider.

    Method query_params must be overridden by derived classes.

    """

    @property
    @abstractmethod
    def query_params(self):
        """
        Return dict of query parameters.

        These parameters are sourced in the object
        implementing this method.

        Returns
        -------
            dict -- dictionary of query parameter values.

        """
        return {}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Morder dataset browser."""
from pprint import pformat
from typing import Any, Dict, Iterable

import ipywidgets as widgets
import pandas as pd
from IPython.display import display, HTML

from ..._version import VERSION
from ..drivers.mordor_driver import (
    MordorDriver,
    MordorEntry,
    download_mdr_file,
    search_mdr_data,
)

__version__ = VERSION
__author__ = "Ian Hellen"


class MordorBrowser:
    """Mordor browser widget."""

    def __init__(self, save_folder: str = ".", use_cached: bool = True):
        """
        Initialize MordorBrowser control.

        Parameters
        ----------
        save_folder : str, optional
            Folder to save downloaded files, by default "."
        use_cached : bool, optional
            If true, downloaded files are not deleted after
            download and are used as a local cache, by default True

        """
        self._save_folder = save_folder
        self._use_cached = use_cached
        self.mordor_driver = MordorDriver()
        self.mordor_driver.connect()
        self.mdr_metadata = self.mordor_driver.mordor_data
        self._df_disp = None

        self.layouts = self._create_layouts()
        self.w_style = {
            "description_width": "150px",
            "font_family": "arial, sans-serif",
        }

        self.widgets: Dict[str, Any] = {}
        self._init_field_ctls()
        self._init_select_dataset()
        self._init_filter_ctrls()

        wgt_title = widgets.HTML(
            value="<h2>Mordor dataset browser</h2>", style={"fontsize": "x-large"}
        )

        browse_ctrls = widgets.VBox(
            [wgt_title, self.widgets["ds_select"], self.widgets["filter_grp"]],
            layout=self.layouts["box_layout"],
        )
        fields_ctrls = widgets.VBox(
            list(self.fields.values()), layout=self.layouts["box_layout"]
        )

        self.datasets: Dict[str, pd.DataFrame] = {}
        self.current_dataset: pd.DataFrame = None
        display(widgets.VBox([browse_ctrls, fields_ctrls]))
        self._df_disp = display(HTML("<p>"), display_id=True)

    @property
    def fields(self):
        """Return set of fields widget controls."""
        return self.widgets["fields"]

    @property
    def selected_dset(self):
        """Return the ID of the selected data set."""
        return self.widgets["ds_select"].value

    def _clear_df_display(self):
        if self._df_disp:
            self._df_disp.update(HTML("<p>"))

    @staticmethod
    def _create_layouts():
        """Set up the widget layouts."""
        return {
            "layout_norm": widgets.Layout(width="70%"),
            "layout_high": widgets.Layout(width="70%", height="200px"),
            "box_layout": widgets.Layout(
                width="80%", border="1px solid", margin="5px", padding="10px"
            ),
            "mitre_select_layout": widgets.Layout(width="40%", height="100px"),
        }

    def _init_filter_ctrls(self):
        """Initialize the filter controls."""
        # text_filter
        self.widgets["filter_text"] = widgets.Text(
            description="Filter", layout=self.layouts["layout_norm"], style=self.w_style
        )
        self.widgets["filter_text"].on_submit(self._update_select_list)
        self.widgets["filter_help"] = widgets.Label(
            value=" comma ORs values, '+' ANDs values"
        )

        # Mitre filters
        self.widgets["sel_techniques"] = widgets.SelectMultiple(
            description="Mitre Techniques",
            options=self._get_mitre_filter_options(
                self.mordor_driver.mdr_idx_tech, self.mordor_driver.mitre_techniques
            ),
            layout=self.layouts["mitre_select_layout"],
            style=self.w_style,
        )

        self.widgets["sel_tactics"] = widgets.SelectMultiple(
            description="Mitre Tactics",
            options=self._get_mitre_filter_options(
                self.mordor_driver.mdr_idx_tact, self.mordor_driver.mitre_tactics
            ),
            layout=self.layouts["mitre_select_layout"],
            style=self.w_style,
        )
        self._reset_filters()
        self.widgets["sel_techniques"].observe(self._update_select_list, names="value")
        self.widgets["sel_tactics"].observe(self._update_select_list, names="value")

        self.widgets["filter_reset"] = widgets.Button(description="Reset filter")
        self.widgets["filter_reset"].on_click(self._reset_filters)
        wgt_filter_grp = widgets.VBox(
            [
                widgets.HBox(
                    [self.widgets["filter_text"], self.widgets["filter_help"]]
                ),
                widgets.HBox(
                    [
                        self.widgets["sel_techniques"],
                        self.widgets["sel_tactics"],
                        self.widgets["filter_reset"],
                    ]
                ),
            ]
        )
        self.widgets["filter_grp"] = widgets.Accordion(children=[wgt_filter_grp])
        self.widgets["filter_grp"].set_title(0, "Filters")
        self.widgets["filter_grp"].selected_index = None

    def _init_select_dataset(self):
        """Initialize the select dataset control."""
        ds_select = widgets.Select(
            description="Data sets",
            options=self._get_md_select_options(),
            layout=self.layouts["layout_norm"],
            style=self.w_style,
        )
        ds_select.observe(self._select_ds_item, names="value")
        self._select_ds_item({"new": next(iter(self.mdr_metadata.keys()))})
        self.widgets["ds_select"] = ds_select

    def _init_field_ctls(self):
        """Initialize the data field controls."""
        fields = {}
        dl_button = None
        for field, field_attrs in MORDOR_FIELDS.items():
            if field == "file_paths":
                dl_button = widgets.Button(description="Download")
                fields[field] = widgets.HBox(
                    [
                        widgets.Select(
                            description=field,
                            layout=self.layouts["layout_norm"],
                            style=self.w_style,
                        ),
                        dl_button,
                    ]
                )
            elif field == "simulation":
                fields[field] = field_attrs["widget"](
                    description=field,
                    layout=self.layouts["layout_high"],
                    style=self.w_style,
                )
            else:
                fields[field] = field_attrs["widget"](
                    description=field,
                    layout=self.layouts["layout_norm"],
                    style=self.w_style,
                )
        self.widgets["fields"] = fields

        if dl_button is not None:
            self.widgets["wgt_dl_button"] = dl_button
            dl_button.on_click(self._download_file)

    def _clear_fields(self):
        """Clear data fields (when nothing is selected)."""
        for field in MORDOR_FIELDS:
            if field == "file_paths":
                self.fields[field].children[0].options = []
            else:
                self.fields[field].value = ""
        self._clear_df_display()

    def _select_ds_item(self, change):  # noqa: MC0001
        """Handle change of dataset selection."""
        item_id = change.get("new")
        mdr_item = self.mdr_metadata.get(item_id)
        if not mdr_item:
            self._clear_fields()
            return

        for field, field_attrs in MORDOR_FIELDS.items():
            if mdr_item and field_attrs["type"] != "cust":
                value = getattr(mdr_item, field)
            else:
                value = None

            if field_attrs["type"] == "text":
                self.fields[field].value = str(value)
            elif field_attrs["type"] == "list":
                self.fields[field].value = ", ".join([str(item) for item in value])
            elif field_attrs["type"] == "raw":
                self.fields[field].value = pformat(value)
            elif field == "attacks":
                field_data = mdr_item.get_attacks()
                self.fields[field].value = _format_attacks(field_data)
            elif field == "file_paths":
                file_paths = mdr_item.get_file_paths()
                self.fields[field].children[0].options = _format_files(file_paths)
            elif field == "notebooks":
                self.fields[field].value = _format_notebooks(mdr_item.notebooks)
            elif field == "query name":
                self.fields[field].value = _format_queries(mdr_item)

        self._clear_df_display()

    def _update_select_list(self, event=None):
        """Update the dataset selection list based on filters."""
        del event
        filtered_tech = set()
        for t_id in self.widgets["sel_techniques"].value:
            filtered_tech.update(self.mordor_driver.mdr_idx_tech[t_id])

        filtered_tact = set()
        for t_id in self.widgets["sel_tactics"].value:
            filtered_tact.update(self.mordor_driver.mdr_idx_tact[t_id])

        md_items_filtered = filtered_tech & filtered_tact
        md_ids = search_mdr_data(
            self.mdr_metadata, self.widgets["filter_text"].value, md_items_filtered
        )
        self.widgets["ds_select"].options = self._get_md_select_options(md_ids)

    def _reset_filters(self, event=None):
        """Reset filter controls to default."""
        del event
        self.widgets["sel_techniques"].value = [
            opt[1] for opt in self.widgets["sel_techniques"].options
        ]
        self.widgets["sel_tactics"].value = [
            opt[1] for opt in self.widgets["sel_tactics"].options
        ]
        self.widgets["filter_text"].value = ""
        self._update_select_list()

    def _get_md_select_options(self, subset: Iterable[str] = None):
        """Return current set of datasets for select control."""
        return [
            (f"{mdr.id} {mdr.title} ({mdr.platform})", mdr.id)
            for mdr in self.mdr_metadata.values()
            if subset is None or mdr.id in subset
        ]

    def _download_file(self, event):
        """Handle download file event."""
        del event
        selection = self.fields["file_paths"].children[0].value
        if selection not in self.datasets:
            result_df = download_mdr_file(
                selection, use_cached=self._use_cached, save_folder=self._save_folder
            )
            if not isinstance(result_df, pd.DataFrame) or result_df.empty:
                result_df = HTML("Could not extract data from this file")
            self.datasets[selection] = result_df
        self.current_dataset = self.datasets[selection]
        self._df_disp.update(self.datasets[selection])

    @staticmethod
    def _get_mitre_filter_options(mordor_index: Dict[str, MordorEntry], mitre_data):
        return [
            (f"{m_id} - {mitre_data.loc[m_id].Name}", m_id) for m_id in mordor_index
        ]


_FMT_STYLE = "border: 1px solid #AAAAAA; padding: 5px"


def _format_attacks(attacks):
    """Format the Mitre Attack data for display."""
    html_text = []
    for attack in attacks:
        sub_tech = f" (sub: {attack.sub_technique}) " if attack.sub_technique else ""
        html_tech = [
            f"<p>Mitre Technique <b>{attack.technique}</b> ",
            sub_tech,
            f": <a href='{attack.technique_uri}' target='_blank'>",
            f"<u>{attack.technique_name}</u></a></p>",
            "Mitre Tactics: ",
        ]

        html_tactics = [
            f"<b>{tactic[0]}</b>: <a href='{tactic[3]}' "
            + f"target='_blank'><u>{tactic[1]}</u></a>"
            for tactic in attack.tactics_full
        ]

        html_tactics = ", ".join(html_tactics)
        html_text.append("".join([*html_tech, html_tactics]))
    content = "<br>".join(html_text)
    return f"<div style='{_FMT_STYLE}'>{content}</span>"


def _format_files(files):
    """Format the Mordor files data for display."""
    return [
        (f"({file['file_type']}) {file['file_path'].split('/')[-1]}", file["file_path"])
        for file in files
    ]


def _format_notebooks(notebooks):
    """Format the Mordor notebooks data for display."""
    if not notebooks:
        return f"<div style='{_FMT_STYLE}'>none</span>"
    nbks = ((nbk["project"], nbk["name"], nbk["link"]) for nbk in notebooks)
    content = "".join(
        [
            f"<p>{nbk[0]} - <a href='{nbk[2]}' target='_blank'><u>{nbk[1]}</u></a></p>"
            for nbk in nbks
        ]
    )
    return f"<div style='{_FMT_STYLE}'>{content}</span>"


def _format_queries(mdr_entry):
    return "\n".join(
        [
            f"{file_path['qry_path']} ({mdr_entry.title}, {mdr_entry.id})"
            for file_path in mdr_entry.get_file_paths()
        ]
    )


MORDOR_FIELDS = {
    "title": {"type": "text", "widget": widgets.Text},
    "id": {"type": "text", "widget": widgets.Text},
    "author": {"type": "text", "widget": widgets.Text},
    "creation_date": {"type": "text", "widget": widgets.Text},
    "modification_date": {"type": "text", "widget": widgets.Text},
    "platform": {"type": "text", "widget": widgets.Text},
    "description": {"type": "text", "widget": widgets.Textarea},
    "tags": {"type": "list", "widget": widgets.Text},
    "file_paths": {
        "type": "cust",
        "widget": widgets.Select,
        "method": "get_file_paths",
        "formatter": _format_files,
    },
    "attacks": {
        "type": "cust",
        "widget": widgets.HTML,
        "method": "get_attacks",
        "formatter": _format_attacks,
    },
    "notebooks": {"type": "cust", "widget": widgets.HTML},
    "simulation": {"type": "raw", "widget": widgets.Textarea},
    "references": {"type": "raw", "widget": widgets.Textarea},
    "query name": {"type": "cust", "widget": widgets.Textarea},
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Setup script for msticpy."""

import re
import setuptools


with open("README.md", "r") as fh:
    LONG_DESC = fh.read()

# pylint: disable=locally-disabled, invalid-name
with open("msticpy/_version.py", "r") as fd:
    v_match = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    __version__ = v_match.group(1) if v_match else "no version"
# pylint: enable=locally-disabled, invalid-name

with open("requirements.txt", "r") as fh:
    INSTALL_REQUIRES = fh.readlines()

with open("requirements-dev.txt", "r") as fh:
    INSTALL_DEV_REQUIRES = fh.readlines()

# Extras definitions
EXTRAS = {
    "dev": INSTALL_DEV_REQUIRES,
    "vt3": ["vt-py>=0.5.4", "vt-graph-api>=1.0.1", "nest_asyncio>=1.4.0"],
    "splunk": ["splunk-sdk>=1.6.0"],
}

setuptools.setup(
    name="msticpy",
    version=__version__,
    author="Ian Hellen",
    author_email="ianhelle@microsoft.com",
    description="MSTIC Security Tools",
    license="MIT License",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/msticpy",
    project_urls={
        "Documentation": "https://msticpy.readthedocs.io",
        "Code": "https://github.com/microsoft/msticpy",
    },
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests", "tests.*", "*.tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS,
    keywords=[
        "security",
        "azure",
        "sentinel",
        "mstic",
        "cybersec",
        "infosec",
        "cyber",
        "cybersecurity",
    ],
    zip_safe=False,
    include_package_data=True,
)
